import logging
import os
import sys

import pytest
import scanpy as sc
from gtars.utils import write_tokens_to_gtok
from tqdm import tqdm

from geniml.region2vec.utils import Region2VecDataset
from geniml.scembed.main import ScEmbed
from geniml.tokenization.main import AnnDataTokenizer

# add parent directory to path
sys.path.append("../")


# set to DEBUG to see more info
logging.basicConfig(level=logging.INFO)


@pytest.fixture
def universe_file():
    return "tests/data/universe.bed"


@pytest.fixture
def hf_model():
    return "databio/r2v-ChIP-atlas-hg38-v2"


@pytest.fixture
def pbmc_data():
    return sc.read_h5ad("tests/data/pbmc_hg38.h5ad")


@pytest.fixture
def pbmc_data_backed():
    return sc.read_h5ad("tests/data/pbmc_hg38.h5ad", backed="r")


def test_model_creation():
    model = ScEmbed()
    assert model


def test_model_training(universe_file: str, pbmc_data: sc.AnnData):
    # remove gensim logging
    logging.getLogger("gensim").setLevel(logging.ERROR)
    model = ScEmbed(tokenizer=AnnDataTokenizer(universe_file))  # set to 1 for testing

    dataset = Region2VecDataset("tests/data/gtok_sample/", convert_to_str=True)
    model.train(dataset, epochs=3, min_count=1)

    # keep only columns with values > 0
    pbmc_data = pbmc_data[:, pbmc_data.X.sum(axis=0) > 0]

    assert model.trained


def test_model_train_and_export(universe_file: str):
    # remove gensim logging
    logging.getLogger("gensim").setLevel(logging.ERROR)
    model = ScEmbed(tokenizer=AnnDataTokenizer(universe_file))  # set to 1 for testing

    dataset = Region2VecDataset("tests/data/gtok_sample/", convert_to_str=True)
    model.train(dataset, epochs=3, min_count=1)

    assert model.trained

    # save
    try:
        model.export("tests/data/model-tests")
        model = ScEmbed.from_pretrained("tests/data/model-tests")

        # ensure model is still trained and has region2vec
        assert model.trained

    finally:
        os.remove("tests/data/model-tests/checkpoint.pt")
        os.remove("tests/data/model-tests/universe.bed")
        os.remove("tests/data/model-tests/config.yaml")


@pytest.mark.skip(reason="Need to get a pretrained model first")
def test_pretrained_scembed_model(hf_model: str, pbmc_data: sc.AnnData):
    model = ScEmbed(hf_model)
    embeddings = model.encode(pbmc_data)
    assert embeddings.shape[0] == pbmc_data.shape[0]


@pytest.mark.skip(reason="This is for my own testing")
def test_end_to_end_training():
    import matplotlib.pyplot as plt
    from umap import UMAP

    data_path = os.path.expandvars("$HOME/Desktop/buenrostro2018.h5ad")
    tokens_path = os.path.expandvars("$HOME/Desktop/tokens")
    universe_path = os.path.expandvars("$HOME/Desktop/universe.bed")
    export_path = os.path.expandvars("$HOME/Desktop/scembed-model")

    # try to remove tokens folder if it exists
    if os.path.exists(tokens_path):
        os.system(f"rm -rf {tokens_path}")

    adata = sc.read_h5ad(data_path)

    tokenizer = AnnDataTokenizer(universe_path)
    tokens = tokenizer.tokenize(adata)

    for i, t in tqdm(enumerate(tokens)):
        file = os.path.join(tokens_path, f"cell{i}.gtok")
        write_tokens_to_gtok(file, t.ids)

    dataset = Region2VecDataset(tokens_path, convert_to_str=True, shuffle=True)

    model = ScEmbed(tokenizer=tokenizer)
    model.train(dataset, epochs=3, num_cpus=5)

    model.export(export_path)

    model2 = ScEmbed.from_pretrained(export_path)
    embeddings = model2.encode(adata)

    reducer = UMAP(n_components=2, random_state=42)
    embedding = reducer.fit_transform(embeddings)

    fig, ax = plt.subplots(figsize=(10, 10))
    ax.scatter(embedding[:, 0], embedding[:, 1], s=0.1)
    fig.savefig("umap_test.png")



import json
from typing import Dict, List, Union

from huggingface_hub import hf_hub_download

from .backends import HNSWBackend
from .const import HF_INDEX, HF_METADATA, HF_PAYLOADS
from .interfaces import Text2BEDSearchInterface
from .query2vec import Text2Vec


def anecdotal_search_from_hf_data(
    query: str, dataset_repo: str, search_model_repo: str, text_embed_model_repo: str, k: int = 10
) -> List[Dict[str, Union[float, int, Dict[str, str]]]]:
    """
    Test the retrieval performance of a trained search model on a dataset on huggingface

    :param query: user input search term
    :param dataset_repo: huggingface repository of the dataset
    :param search_model_repo: huggingface repository of the search model
    :param text_embed_model_repo:huggingface repository of the text encoder model
    :param k: number of returned result
    :return:a dictionary containing search score (distance to the mapped query vector) and
    search result metadata.
    """

    # download embedding vector backends and metadata from huggingface
    index_path = hf_hub_download(dataset_repo, HF_INDEX, repo_type="dataset")
    payloads_path = hf_hub_download(dataset_repo, HF_PAYLOADS, repo_type="dataset")
    metadata_path = hf_hub_download(dataset_repo, HF_METADATA, repo_type="dataset")

    # evaluation backend
    eval_backend = HNSWBackend(local_index_path=index_path, payloads=payloads_path)

    # load metadata
    with open(metadata_path, "r") as f:
        metadata_dict = json.load(f)

    text2vec = Text2Vec(text_embed_model_repo, search_model_repo)
    search_interface = Text2BEDSearchInterface(eval_backend, text2vec)

    search_results = search_interface.query_search(query, k, with_payload=True, with_vectors=False)

    # curate output dictionary
    result_files_id_dict = {
        search_results[i]["payload"]["file"]: i for i in range(len(search_results))
    }
    for attribute in metadata_dict:
        for metadata in metadata_dict[attribute]:
            for file in metadata_dict[attribute][metadata]:
                try:
                    search_results[result_files_id_dict[file]]["payload"][attribute] = metadata
                except:
                    continue

    return search_results



from .backends import HNSWBackend, QdrantBackend
from .filebackend_tools import merge_backends
from .interfaces import BED2BEDSearchInterface, Text2BEDSearchInterface
from .query2vec import BED2Vec, Text2Vec
from .search_eval import anecdotal_search_from_hf_data
from .utils import rand_eval



import os
from logging import getLogger
from typing import List, Union

import numpy as np

try:
    import torch
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )

from gtars.tokenizers import Region as GRegion
from gtars.tokenizers import RegionSet as GRegionSet
from huggingface_hub import hf_hub_download
from rich.progress import track

from ..io import Region, RegionSet
from ..models import ExModel
from ..tokenization.main import Tokenizer, TreeTokenizer
from .const import (
    CONFIG_FILE_NAME,
    DEFAULT_EMBEDDING_DIM,
    DEFAULT_EPOCHS,
    DEFAULT_MIN_COUNT,
    DEFAULT_WINDOW_SIZE,
    MODEL_FILE_NAME,
    MODULE_NAME,
    POOLING_METHOD_KEY,
    POOLING_TYPES,
    UNIVERSE_FILE_NAME,
)
from .models import Region2Vec
from .utils import (
    Region2VecDataset,
    export_region2vec_model,
    load_local_region2vec_model,
    train_region2vec_model,
)

_GENSIM_LOGGER = getLogger("gensim")
_LOGGER = getLogger(MODULE_NAME)

# demote gensim logger to warning
_GENSIM_LOGGER.setLevel("WARNING")


class Region2VecExModel(ExModel):
    def __init__(
        self,
        model_path: str = None,
        tokenizer: TreeTokenizer = None,
        device: str = None,
        pooling_method: POOLING_TYPES = "mean",
        **kwargs,
    ):
        """
        Initialize Region2VecExModel.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param embedding_dim: Dimension of the embedding.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        super().__init__()
        self.model_path: str = model_path
        self.tokenizer: TreeTokenizer
        self.trained: bool = False
        self._model: Region2Vec = None
        self.pooling_method = pooling_method

        if model_path is not None:
            self._init_from_huggingface(model_path)
            self.trained = True

        elif tokenizer is not None:
            self._init_model(tokenizer, **kwargs)

        # set the device
        self._target_device = torch.device(
            device if device else ("cuda" if torch.cuda.is_available() else "cpu")
        )

    def _init_tokenizer(self, tokenizer: Union[TreeTokenizer, str]):
        """
        Initialize the tokenizer.

        :param tokenizer: Tokenizer to initialize.
        """
        if isinstance(tokenizer, str):
            if os.path.exists(tokenizer):
                self.tokenizer = TreeTokenizer(tokenizer)
            else:
                self.tokenizer = TreeTokenizer.from_pretrained(
                    tokenizer
                )  # download from huggingface (or at least try to)
        elif isinstance(tokenizer, TreeTokenizer):
            self.tokenizer = tokenizer
        else:
            raise TypeError("tokenizer must be a path to a bed file or an TreeTokenizer object.")

    def _init_model(self, tokenizer, **kwargs):
        """
        Initialize the core model. This will initialize the model from scratch.

        :param kwargs: Additional keyword arguments to pass to the model.
        """
        self._init_tokenizer(tokenizer)
        self._model = Region2Vec(
            len(self.tokenizer),
            embedding_dim=kwargs.get("embedding_dim", DEFAULT_EMBEDDING_DIM),
        )

    @property
    def model(self):
        """
        Get the core Region2Vec model.
        """
        return self._model

    def add_tokenizer(self, tokenizer: Tokenizer, **kwargs):
        """
        Add a tokenizer to the model. This should be use when the model
        is not initialized with a tokenizer.

        :param tokenizer: Tokenizer to add to the model.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        if self._model is not None:
            raise RuntimeError("Cannot add a tokenizer to a model that is already initialized.")

        self.tokenizer = tokenizer
        if not self.trained:
            self._init_model(**kwargs)

    def _load_local_model(self, model_path: str, vocab_path: str, config_path: str):
        """
        Load the model from a checkpoint.

        :param str model_path: Path to the model checkpoint.
        :param str vocab_path: Path to the vocabulary file.
        """
        _model, config = load_local_region2vec_model(model_path, config_path)
        tokenizer = TreeTokenizer(vocab_path)

        self._model = _model
        self.tokenizer = tokenizer

        self.trained = True
        if POOLING_METHOD_KEY in config:
            self.pooling_method = config[POOLING_METHOD_KEY]

    def _init_from_huggingface(
        self,
        model_path: str,
        model_file_name: str = MODEL_FILE_NAME,
        universe_file_name: str = UNIVERSE_FILE_NAME,
        config_file_name: str = CONFIG_FILE_NAME,
        **kwargs,
    ):
        """
        Initialize the model from a huggingface model. This uses the model path
        to download the necessary files and then "build itself up" from those. This
        includes both the actual model and the tokenizer.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param str model_file_name: Name of the model file.
        :param str universe_file_name: Name of the universe file.
        :param kwargs: Additional keyword arguments to pass to the hf download function.
        """
        model_file_path = hf_hub_download(model_path, model_file_name, **kwargs)
        universe_path = hf_hub_download(model_path, universe_file_name, **kwargs)
        config_path = hf_hub_download(model_path, config_file_name, **kwargs)

        self._load_local_model(model_file_path, universe_path, config_path)

    @classmethod
    def from_pretrained(
        cls,
        path_to_files: str,
        model_file_name: str = MODEL_FILE_NAME,
        universe_file_name: str = UNIVERSE_FILE_NAME,
        config_file_name: str = CONFIG_FILE_NAME,
    ) -> "Region2VecExModel":
        """
        Load the model from a set of files that were exported using the export function.

        :param str path_to_files: Path to the directory containing the files.
        :param str model_file_name: Name of the model file.
        :param str universe_file_name: Name of the universe file.
        """
        model_file_path = os.path.join(path_to_files, model_file_name)
        universe_file_path = os.path.join(path_to_files, universe_file_name)
        config_file_path = os.path.join(path_to_files, config_file_name)

        instance = cls()
        instance._load_local_model(model_file_path, universe_file_path, config_file_path)
        instance.trained = True

        return instance

    def _validate_data_for_training(
        self, data: Union[List[RegionSet], List[str], List[List[Region]]]
    ) -> List[RegionSet]:
        """
        Validate the data for training. This will return a list of RegionSets if the data is valid.

        :param Union[List[RegionSet], List[str]] data: List of data to train on. This is either
                                                       a list of RegionSets or a list of paths to bed files.
        :return: List of RegionSets.
        """
        if not isinstance(data, list):
            raise TypeError("data must be a list or RegionSets or a list of paths to bed files.")
        if len(data) == 0:
            raise ValueError("data must not be empty.")

        # check if the data is a list of RegionSets
        if isinstance(data[0], RegionSet):
            return data
        elif isinstance(data[0], str):
            return [RegionSet(f) for f in data]
        elif isinstance(data[0], list) and isinstance(data[0][0], Region):
            return [RegionSet([r for r in region_list]) for region_list in data]

    def train(
        self,
        dataset: Region2VecDataset,
        window_size: int = DEFAULT_WINDOW_SIZE,
        epochs: int = DEFAULT_EPOCHS,
        min_count: int = DEFAULT_MIN_COUNT,
        num_cpus: int = 1,
        seed: int = 42,
        save_checkpoint_path: str = None,
        gensim_params: dict = {},
        load_from_checkpoint: str = None,
    ) -> bool:
        """
        Train the model.

        :param dataset Region2VecDataset: Dataset to train on.
        :param int window_size: Window size for the model.
        :param int epochs: Number of epochs to train for.
        :param int min_count: Minimum count for a region to be included in the vocabulary.
        :param int num_cpus: Number of cpus to use for training.
        :param int seed: Seed to use for training.
        :param str save_checkpoint_path: Path to save the model checkpoints to.
        :param dict gensim_params: Additional parameters to pass to the gensim model.
        :param str load_from_checkpoint: Path to a checkpoint to load from.

        :return np.ndarray: Loss values for each epoch.
        """
        # validate a model exists
        if self._model is None:
            raise RuntimeError(
                "Cannot train a model that has not been initialized. Please initialize the model first using a tokenizer or from a huggingface model."
            )

        gensim_model = train_region2vec_model(
            dataset,
            embedding_dim=self._model.embedding_dim,
            window_size=window_size,
            epochs=epochs,
            min_count=min_count,
            num_cpus=num_cpus,
            seed=seed,
            save_checkpoint_path=save_checkpoint_path,
            gensim_params=gensim_params,
            load_from_checkpoint=load_from_checkpoint,
        )

        # once done training, set the weights of the pytorch model in self._model
        for id in track(
            gensim_model.wv.key_to_index,
            total=len(gensim_model.wv.key_to_index),
            description="Setting weights.",
        ):
            self._model.projection.weight.data[int(id)] = torch.tensor(gensim_model.wv[id])

        # set the model as trained
        self.trained = True

        return True

    def export(
        self,
        path: str,
        checkpoint_file: str = MODEL_FILE_NAME,
        universe_file: str = UNIVERSE_FILE_NAME,
        config_file: str = CONFIG_FILE_NAME,
    ):
        """
        Function to facilitate exporting the model in a way that can
        be directly uploaded to huggingface. This exports the model
        weights and the vocabulary.

        :param str path: Path to export the model to.
        """

        export_region2vec_model(
            self._model,
            self.tokenizer,
            path,
            checkpoint_file=checkpoint_file,
            universe_file=universe_file,
            config_file=config_file,
        )

    def encode(
        self,
        regions: Union[str, Region, List[Region], RegionSet, GRegionSet],
        pooling: POOLING_TYPES = None,
    ) -> np.ndarray:
        """
        Get the vector for a region.

        :param regions: Region to get the vector for.
        :param pooling: Pooling type to use.

        :return np.ndarray: Vector for the region.
        """
        # allow for overriding the pooling method
        pooling = pooling or self.pooling_method

        # data validation
        if isinstance(regions, Region):
            regions = [regions]
        if isinstance(regions, str):
            regions = RegionSet(regions)
        if not isinstance(regions[0], Region) and not isinstance(regions[0], GRegion):
            raise TypeError("regions must be a list of Region objects.")

        if pooling not in ["mean", "max"]:
            raise ValueError(f"pooling must be one of {POOLING_TYPES}")

        # tokenize the regionm -- need to pass it as a list because the tokenizer expects a list
        tokens = [self.tokenizer([r]) for r in regions]
        token_tensors = [
            torch.tensor(token_set.to_ids(), dtype=torch.long) for token_set in tokens
        ]

        region_embeddings = []

        for token_tensor in token_tensors:
            if pooling == "mean":
                region_embeddings.append(
                    torch.mean(self._model.projection(token_tensor), axis=0).detach().numpy()
                )
            elif pooling == "max":
                region_embeddings.append(
                    torch.max(self._model.projection(token_tensor), axis=0).detach().numpy()
                )
            else:
                # this should be unreachable
                raise ValueError(f"pooling must be one of {POOLING_TYPES}")

        return np.vstack(region_embeddings)



from collections import Counter

import scanpy as sc
from qdrant_client import QdrantClient
from rich.progress import track


class AnnotationServer(QdrantClient):
    def __init__(
        self,
        location: str = None,
        url: str = None,
        port: int = None,
        api_key: str = None,
        collection_name: str = None,
        timeout: float = 10,
    ):
        """
        A class for querying a Qdrant server for cell type predictions. This class requires that you have
        a Qdrant server running with a collection of cell type embeddings. You can create this collection using
        the `geniml/examples/scembed/load_qdrant.ipynb` script.

        :param str collection_name: The name of the collection to query.
        :param str location: The location of the Qdrant server. This should be in the format `host:port`.
        :param str url: The URL of the Qdrant server.
        :param int port: The port of the Qdrant server.
        :param str api_key: The API key for the Qdrant server.
        :param float timeout: The timeout for the Qdrant server.
        """
        super().__init__(location, url=url, port=port, api_key=api_key, timeout=timeout)
        self.collection_name = collection_name


class Annotator:
    def __init__(
        self,
        collection_name: str = None,
        location: str = None,
        url: str = None,
        port: int = None,
        timeout: float = 10,
    ):
        """
        A class for annotating single cell data with cell type predictions. This class requires that you have
        a Qdrant server running with a collection of cell type embeddings. You can create this collection using
        the `geniml/examples/scembed/load_qdrant.ipynb` script.

        :param str collection_name: The name of the collection to query.
        :param str location: The location of the Qdrant server. This should be in the format `host:port`.
        :param str url: The URL of the Qdrant server.
        :param int port: The port of the Qdrant server.
        :param float timeout: The timeout for the Qdrant server.

        """
        self.collection_name = collection_name
        self.url = url
        self.port = port
        self.timeout = timeout
        self._annotation_server = AnnotationServer(
            location=location,
            url=self.url,
            port=self.port,
            collection_name=self.collection_name,
            timeout=self.timeout,
        )

    def annotate(
        self,
        adata: sc.AnnData,
        embedding_key: str = "embedding",
        key_added: str = "pred_celltype",
        cluter_key: str = "leiden",
        knn: int = 3,
        score_threshold: float = 0.5,
    ):
        """
        Annotate a sc.AnnData object with cell type predictions for each cluster. This functions requires
        that you have 1) clustered your data and 2) embedded your data using some pretrained model (databio/multiome).

        You can cluster your data using the `scanpy`

        It is *imperative* that the model used to embed your single cells is the same model used to produce the
        embeddings in the database. Otherwise, the predictions will not be accurate; in fact they will be meaningless.

        :param sc.AnnData adata: The annotated data.
        :param str embedding_key: The key in `adata.obsm` where the embeddings are stored.
        :param str key_added: The key in `adata.obs` where the cell type predictions will be stored.
        :param str cluter_key: The key in `adata.obs` where the cluster labels are stored.
        :param int knn: The number of nearest neighbors to use when querying the database.
        :param float score_threshold: The score threshold to use when querying the database.
        """

        _temp_key = "putative_cell_type"

        # check that the embedding key exists
        if embedding_key not in adata.obsm.keys():
            raise ValueError(
                f"Embedding key '{embedding_key}' not found in adata.obsm. Please embed your data first."
            )
        if cluter_key not in adata.obs.keys():
            raise ValueError(
                f"Cluster key '{cluter_key}' not found in adata.obs. Please cluster your data first."
            )

        # init list
        scembed_cell_type_preds = []

        # use qdrant and a simple KNN approach to attach cell types to the embeddings
        for embedding in track(adata.obsm["embedding"], total=len(adata.obsm["embedding"])):
            results = self._annotation_server.search(
                collection_name=self.collection_name,
                query_vector=embedding,
                limit=knn,
                score_threshold=score_threshold,
            )

            result_dicts = [result.dict() for result in results]

            # count "cell_type" in all dicts
            c = Counter([result["payload"]["cell_type"] for result in result_dicts])

            # simply get the name of the top most common and thats it
            try:
                cell_type = c.most_common(1)[0][0]
            except IndexError:
                cell_type = "Unknown"

            scembed_cell_type_preds.append(cell_type)

        # add to adata, these are just putative cell types
        # we will take a consensus vote later using clusters
        adata.obs[_temp_key] = scembed_cell_type_preds

        # now take a consensus vote for each cluster
        cluster_celltypes = {}
        for cluster in adata.obs["leiden"].unique():
            cluster_celltypes[cluster] = Counter(
                adata.obs[adata.obs["leiden"] == cluster][_temp_key]
            ).most_common(1)[0][0]

        # map the cluster_to_cell_type dictionary to the leiden column
        adata.obs[key_added] = adata.obs[cluter_key].map(cluster_celltypes)



from typing import Tuple

import lightning as L
import torch
import torch.nn as nn
from lightning.pytorch.utilities.types import OptimizerLRScheduler
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

from ..atacformer.main import AtacformerExModel


class CellTypeFineTuneAdapter(L.LightningModule):
    """
    An adapter for fine-tuning a model on cell type classification.
    """

    def __init__(
        self,
        model: AtacformerExModel,
        num_classes: int,
        init_lr: float = 1e-5,
        adamw_weight_decay: float = 0.01,
        **kwargs,
    ):
        """
        Instantiate a fine-tuning trainer.

        :param Region2VecExModel | scEmbed | Atacformer model: The model to fine-tune.

        :param kwargs: Additional arguments to pass to the LightningModule constructor.
        """
        super().__init__(**kwargs)

        self.r2v_model = model._model
        self.tokenizer = model.tokenizer
        self.loss_fn = torch.nn.CrossEntropyLoss()
        self.cos_sim = nn.CosineSimilarity()
        self.ffn = nn.Linear(model._model.d_model * 3, num_classes)
        self._exmodel = model
        self.init_lr = init_lr
        self.adamw_weight_decay = adamw_weight_decay

    def mean_pool(token_embeds: torch.Tensor, attention_mask: torch.Tensor):
        """
        Perform mean-pooling on the token embeddings.

        From: https://www.pinecone.io/learn/series/nlp/train-sentence-transformers-softmax/
        """
        # reshape attention_mask to cover 768-dimension embeddings
        in_mask = attention_mask.unsqueeze(-1).expand(token_embeds.size()).float()
        # perform mean-pooling but exclude padding tokens (specified by in_mask)
        pool = torch.sum(token_embeds * in_mask, 1) / torch.clamp(in_mask.sum(1), min=1e-9)
        return pool

    def forward(self, x, mask=None):
        return self.r2v_model(x, mask=mask)

    def compute_loss(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        """
        Compute the loss for the batch.

        :param batch: The batch
        :param batch_idx: The batch index

        :return: The loss
        """
        cell1, cell2, target, attn1, attn2 = batch

        # forward pass for the batch
        u = self.r2v_model(cell1, mask=attn1)
        v = self.r2v_model(cell2, mask=attn2)

        # pool the embeddings using mean
        u = torch.mean(u, dim=1)
        v = torch.mean(v, dim=1)

        uv = torch.sub(u, v)
        uv_abs = torch.abs(uv)

        # concatenate the embeddings
        uv_concat = torch.cat((u, v, uv_abs), dim=1)

        # pass through the linear layer
        output = self.ffn(uv_concat)

        # compute the loss
        loss = self.loss_fn(output, target)

        return loss

    def training_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        """
        Perform a training step.

        :param batch: The batch
        :param batch_idx: The batch index

        :return: The loss
        """
        loss = self.compute_loss(batch, batch_idx)
        self.log("train_loss", loss)
        return loss

    def on_validation_start(self) -> None:
        """
        Perform any setup before validation starts.

        This needs to be here because otherwise the
        TransformerEncoder layers fail (for some reason).
        """
        self.eval()
        torch.set_grad_enabled(True)

    def validation_step(self, batch, batch_idx):
        """
        Perform a validation step.

        :param batch: The batch
        :param batch_idx: The batch index

        :return: The loss
        """
        loss = self.compute_loss(batch, batch_idx)
        self.log("val_loss", loss)
        return loss

    def configure_optimizers(self):
        init_lr = self.init_lr or 1e-6
        optimizer = torch.optim.AdamW(
            self.parameters(), lr=init_lr, weight_decay=self.adamw_weight_decay
        )

        self.trainer.fit_loop.setup_data()

        total_steps = len(self.trainer.train_dataloader) * self.trainer.max_epochs

        scheduler = {
            "scheduler": CosineAnnealingWarmRestarts(
                optimizer,
                T_0=total_steps // 10,  # Restart every 1/10th of total steps
                T_mult=1,  # Keep the same cycle length
                eta_min=1e-7,  # Minimum learning rate
            ),
            "interval": "step",
            "frequency": 1,
        }

        # return optimizer
        return {"optimizer": optimizer, "lr_scheduler": scheduler}


class MLMAdapter(L.LightningModule):
    """
    An adapter for training Atacformer on a masked language modeling task.
    """

    def __init__(self, model: AtacformerExModel, **kwargs):
        """
        Instantiate the masked-language-modeling adapter.

        :param Atacformer model: The model to train.
        :param int pad_token_id: The token to use for padding
        :param int mask_token_id: The token to use for masking
        :param kwargs: Additional arguments to pass to the LightningModule constructor.
        """
        super().__init__(**kwargs)
        self.loss_fn = nn.CrossEntropyLoss()
        # linear layer acts as a classification layer for training
        # but is not used during inference
        self.linear = nn.Linear(model._model.d_model, model._model.vocab_size)
        self.r2v_model = model._model
        self.tokenizer = model.tokenizer
        self.init_lr = kwargs.get("init_lr", 1e-5)

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        token_embeddings = self.r2v_model(x, mask=mask)
        logits = self.linear(token_embeddings)
        return logits

    def configure_optimizers(self) -> OptimizerLRScheduler:
        """
        We use the AdamW optimizer with a learning rate of 1e-6.

        See here: https://arxiv.org/abs/2302.01107

        > By default, AdamW [62], a variant of Adam which decouples the L2 regularization and the
        weight decay, is the most widely used optimizer for Transformers.
        """
        init_lr = self.init_lr or 1e-6
        optimizer = torch.optim.AdamW(self.parameters(), lr=init_lr, weight_decay=0.01)

        self.trainer.fit_loop.setup_data()

        total_steps = len(self.trainer.train_dataloader) * self.trainer.max_epochs

        scheduler = {
            "scheduler": CosineAnnealingWarmRestarts(
                optimizer,
                T_0=total_steps // 10,  # Restart every 1/10th of total steps
                T_mult=1,  # Keep the same cycle length
                eta_min=1e-7,  # Minimum learning rate
            ),
            "interval": "step",
            "frequency": 1,
        }

        # return optimizer
        return {"optimizer": optimizer, "lr_scheduler": scheduler}

    def compute_loss(
        self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], batch_idx: int
    ):
        """
        Compute the loss for the batch.

        :param batch: The batch
        :param batch_idx: The batch index

        :return: The loss
        """
        # move the batch to the device
        tokens, masked_tokens, masked_token_indexes, attention_mask = batch

        # forward pass for the batch
        output = self.forward(masked_tokens, mask=attention_mask)

        # get predictions and targets
        # the predictions are the logits for the masked tokens
        # defined by the masked_token_indexes
        predictions = output.view(-1, self.r2v_model.vocab_size)[masked_token_indexes]
        targets = tokens.view(-1)[masked_token_indexes]

        # reshape once more
        predictions = predictions.view(predictions.shape[0] * predictions.shape[1], -1)
        targets = targets.view(targets.shape[0] * targets.shape[1])

        # compute the loss
        loss = self.loss_fn(predictions, targets)

        return loss

    def training_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        """
        Perform a training step.

        The batch is a tuple of (tokens, masked_tokens, mask_ids). This step performs
        masked language modeling as described in the original BERT paper (https://arxiv.org/abs/1810.04805).

        :param batch: The batch
        :param batch_idx: The batch index

        """
        loss = self.compute_loss(batch, batch_idx)
        self.log("train_loss", loss)
        return loss

    # this breaks everything -- and I have NO idea why...
    def on_validation_start(self) -> None:
        """
        Perform any setup before validation starts.

        This needs to be here because otherwise the
        TransformerEncoder layers fail (for some reason).
        """
        self.eval()
        torch.set_grad_enabled(True)

    def validation_step(
        self,
        batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        batch_idx: int,
    ):
        """
        Perform a training step.

        The batch is a tuple of (tokens, masked_tokens, mask_ids). This step performs
        masked language modeling as described in the original BERT paper (https://arxiv.org/abs/1810.04805).

        :param batch: The batch
        :param batch_idx: The batch index

        """
        loss = self.compute_loss(batch, batch_idx)
        self.log("val_loss", loss)
        return loss



import logging
import os
import sys

import yaml


class BedshiftYAMLHandler(object):
    def __init__(self, bedshifter, yaml_fp, logger=None):
        """
        Handles Bedshift perturbations from yaml files

        :param bedshift.Bedshift bedshifter: a Bedshift object
        :param str yaml_fp: path to yaml file
        :param logging.logger logger: logger object
        """
        self.bedshifter = bedshifter
        self.yaml_fp = yaml_fp
        if logger is not None:
            self._LOGGER = logger
        else:
            self._LOGGER = logging.getLogger("BedshiftYAMLHandler")

    def _print_sample_config(self):
        """
        bedshift_operations:
          - add:
            rate: 0.1
            mean: 100
            stdev: 20
          - drop_from_file:
            file: tests/test.bed
            rate: 0.1
            delimiter: \t
          - shift_from_file:
            file: bedshifted_test.bed
            rate: 0.3
            mean: 100
            stdev: 200
          - add_from_file:
            file: tests/small_test.bed
            rate: 0.2
          - cut:
            rate: 0.2
          - drop:
            rate: 0.30
          - shift:
            rate: 0.05
            mean: 100
            stdev: 200
          - merge:
            rate: 0.15
        """
        self._LOGGER.info(self._print_sample_config.__doc__)

    def _read_from_yaml(self, fp):
        with open(fp, "r") as yaml_file:
            config_data = yaml.load(yaml_file, Loader=yaml.FullLoader)
        self._LOGGER.info("Loaded configuration settings from {}".format(fp))
        return config_data

    def handle_yaml(self):
        """
        Performs perturbations provided in the yaml config file in the order they were provided.
        """
        data = self._read_from_yaml(self.yaml_fp)
        operations = [operation for operation in data["bedshift_operations"]]
        num_changed = 0

        for operation in operations:
            ##### add #####
            if set(["add", "rate", "mean", "stdev"]) == set(list(operation.keys())):
                rate = operation["rate"]
                mean = operation["mean"]
                std = operation["stdev"]
                num_added = self.bedshifter.add(rate, mean, std)
                num_changed += num_added

            ##### add_from_file with no delimiter provided #####
            elif set(["add_from_file", "file", "rate"]) == set(list(operation.keys())):
                fp = operation["file"]
                if os.path.isfile(fp):
                    add_rate = operation["rate"]
                    num_added = self.bedshifter.add_from_file(fp, add_rate)
                    num_changed += num_added
                else:
                    self._logger.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### add_from_file with delimiter provided #####
            elif set(["add_from_file", "file", "rate", "delimiter"]) == set(
                list(operation.keys())
            ):
                fp = operation["file"]
                if os.path.isfile(fp):
                    add_rate = operation["rate"]
                    delimiter = operation["delimiter"]
                    num_added = self.bedshifter.add_from_file(fp, add_rate, delimiter)
                    num_changed += num_added
                else:
                    self._logger.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### drop #####
            elif set(["drop", "rate"]) == set(list(operation.keys())):
                rate = operation["rate"]
                num_dropped = self.bedshifter.drop(rate)
                num_changed += num_dropped

            ##### drop_from_file with no delimiter provided #####
            elif set(["drop_from_file", "file", "rate"]) == set(list(operation.keys())):
                fp = operation["file"]
                if os.path.isfile(fp):
                    drop_rate = operation["rate"]
                    num_dropped = self.bedshifter.drop_from_file(fp, drop_rate)
                    num_changed += num_dropped
                else:
                    self._LOGGER.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### drop_from_file with delimiter provided #####
            elif set(["drop_from_file", "file", "rate", "delimiter"]) == set(
                list(operation.keys())
            ):
                fp = operation["file"]
                if os.path.isfile(fp):
                    drop_rate = operation["rate"]
                    delimiter = operation["delimiter"]
                    num_dropped = self.bedshifter.drop_from_file(fp, drop_rate, delimiter)
                    num_changed += num_dropped
                else:
                    self._LOGGER.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### shift #####
            elif set(["shift", "rate", "mean", "stdev"]) == set(list(operation.keys())):
                rate = operation["rate"]
                mean = operation["mean"]
                std = operation["stdev"]
                num_shifted = self.bedshifter.shift(rate, mean, std)
                num_changed += num_shifted

            ##### shift_from_file #####
            elif set(["shift_from_file", "file", "rate", "mean", "stdev"]) == set(
                list(operation.keys())
            ):
                fp = operation["file"]
                if os.path.isfile(fp):
                    rate = operation["rate"]
                    mean = operation["mean"]
                    std = operation["stdev"]
                    num_shifted = self.bedshifter.shift_from_file(fp, rate, mean, std)
                    num_changed += num_shifted
                else:
                    self._LOGGER.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### shift_from_file with delimiter provided #####
            elif set(["shift_from_file", "file", "rate", "mean", "stdev", "delimiter"]) == set(
                list(operation.keys())
            ):
                fp = operation["file"]
                if os.path.isfile(fp):
                    rate = operation["rate"]
                    mean = operation["mean"]
                    std = operation["stdev"]
                    delimiter = operation["delimiter"]
                    num_shifted = self.bedshifter.shift_from_file(fp, rate, mean, std, delimiter)
                    num_changed += num_shifted
                else:
                    self._LOGGER.error("File '{}' does not exist.".format(fp))
                    sys.exit(1)

            ##### cut #####
            elif set(["cut", "rate"]) == set(list(operation.keys())):
                rate = operation["rate"]
                num_cut = self.bedshifter.cut(rate)
                num_changed += num_cut

            ##### merge #####
            elif set(["merge", "rate"]) == set(list(operation.keys())):
                rate = operation["rate"]
                num_merged = self.bedshifter.merge(rate)
                num_changed += num_merged

            else:
                self._LOGGER.error(
                    "\n\nInvalid settings entered in the config file. Please refer to the example below.\n\n"
                )
                self._print_sample_config()
                sys.exit(1)

        return num_changed



import os
import sys
from typing import Union

import scanpy as sc
import torch
import torch.nn as nn
from huggingface_hub import hf_hub_download
from yaml import safe_dump, safe_load

from ..models.main import ExModel
from ..tokenization.main import AnnDataTokenizer
from .const import (
    CONFIG_FILE_NAME,
    CONTEXT_SIZE_KEY,
    D_FF_KEY,
    D_MODEL_KEY,
    MODEL_FILE_NAME,
    N_HEADS_KEY,
    N_LAYERS_KEY,
    POOLING_METHOD_KEY,
    POOLING_TYPES,
    UNIVERSE_CONFIG_FILE_NAME,
    UNIVERSE_FILE_NAME,
    VOCAB_SIZE_KEY,
)

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


class Atacformer(nn.Module):
    def __init__(
        self,
        vocab_size: int,
        d_model: int = 768,
        n_heads: int = 12,
        n_layers: int = 12,
        d_ff: int = 3072,
    ):
        """
        Atacformer is a transformer-based model for ATAC-seq data. It closely follows
        the architecture of BERT, but with a few modifications:
        - positional embeddings set to 0 since ATAC-seq data is not sequential
        - no next sentence prediction task
        """
        super().__init__()
        self.d_model = d_model
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.d_ff = d_ff
        self.vocab_size = vocab_size

        # embedding layer
        self.embedding = nn.Embedding(vocab_size, d_model)

        # transformer encoder
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=d_model,
            nhead=n_heads,
            batch_first=True,
            dim_feedforward=d_ff,
            norm_first=True,
        )

        # stack the encoder layers
        self.transformer_encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=n_layers)

    def forward(self, x: torch.Tensor, mask: torch.Tensor = None) -> torch.Tensor:
        """
        :param torch.Tensor x: Input tensor of shape (batch_size, seq_len)
        :param torch.Tensor mask: Mask tensor of shape (batch_size, seq_len)
        :return torch.Tensor: Output tensor of shape (batch_size, seq_len, d_model). I.e. an embedding for each token.
        """
        # get the embeddings
        x = self.embedding(x)

        # set the positional embeddings to 0
        x = x + torch.zeros_like(x)

        # pass through the transformer
        x = self.transformer_encoder(x, src_key_padding_mask=mask)

        return x


class AtacformerExModel(ExModel):
    """
    An "extended model" for Atacformer. This class provides a wrapper around the
    Atacformer model, and provides additional functionality such as tokenization
    and encoding of new data.
    """

    def __init__(
        self,
        model_path: str = None,
        tokenizer: AnnDataTokenizer = None,
        pooling_method: POOLING_TYPES = "mean",
        device: str = None,
        context_size: int = 2048,
        **kwargs,
    ):
        """
        Initialize Region2VecExModel.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param embedding_dim: Dimension of the embedding.
        :param kwargs: Additional keyword arguments to pass to the model.
        """
        super().__init__()
        self.model_path: str = model_path
        self.tokenizer: AnnDataTokenizer
        self.trained: bool = False
        self._model: Atacformer = None
        self.pooling_method = pooling_method
        self.context_size = context_size

        if model_path is not None:
            self._init_from_huggingface(model_path)
            self.trained = True

        elif tokenizer is not None:
            self._init_model(tokenizer, **kwargs)

        # set the device
        self._target_device = torch.device(
            device if device else ("cuda" if torch.cuda.is_available() else "cpu")
        )

        if self._model is not None:
            self._model = self._model.to(self._target_device)

    def _init_tokenizer(self, tokenizer: Union[AnnDataTokenizer, str]):
        """
        Initialize the tokenizer.

        :param tokenizer: Tokenizer to initialize.
        """
        if isinstance(tokenizer, str):
            if os.path.exists(tokenizer):
                self.tokenizer = AnnDataTokenizer(tokenizer)
            else:
                self.tokenizer = AnnDataTokenizer.from_pretrained(
                    tokenizer
                )  # download from huggingface (or at least try to)
        elif isinstance(tokenizer, AnnDataTokenizer):
            self.tokenizer = tokenizer
        else:
            raise TypeError(
                "tokenizer must be a path to a bed file or an `AnnDataTokenizer object."
            )

    def _init_model(self, tokenizer, **kwargs):
        """
        Initialize the core model. This will initialize the model from scratch.

        :param kwargs: Additional keyword arguments to pass to the model.
        """
        self._init_tokenizer(tokenizer)
        self._model = Atacformer(
            len(self.tokenizer),
            d_model=kwargs.get("d_model", 768),
            n_heads=kwargs.get("n_heads", 12),
            n_layers=kwargs.get("n_layers", 12),
            d_ff=kwargs.get("d_ff", 3072),
        )

    @property
    def model(self):
        """
        Get the core Region2Vec model.
        """
        return self._model

    def _load_local_model(self, model_path: str, vocab_path: str, config_path: str):
        """
        Load the model from a checkpoint.

        :param str model_path: Path to the model checkpoint.
        :param str vocab_path: Path to the vocabulary file.
        """

        # check the type of the tokenizer
        # open the toml file and read the `tokenizer_type` key
        with open(vocab_path, "rb") as fp:
            tokenizer_config = tomllib.load(fp)

        tokenizer_type = tokenizer_config.get("tokenizer_type", "tree")

        # init the tokenizer - only one option for now
        self.tokenizer = AnnDataTokenizer(vocab_path, tokenizer_type=tokenizer_type)

        # load the model state dict (weights)
        params = torch.load(model_path)

        # get the model config (vocab size, embedding size)
        with open(config_path, "r") as f:
            config = safe_load(f)

        # try with new key first, then old key for backwards compatibility
        d_model = config.get(D_MODEL_KEY, None)
        if d_model is None:
            raise KeyError(
                f"Could not find embedding dimension in config file. Expected key {D_MODEL_KEY}."
            )

        n_heads = config.get(N_HEADS_KEY, None)
        if n_heads is None:
            raise KeyError(f"Could not find n_heads in config file. Expected key {N_HEADS_KEY}.")

        n_layers = config.get(N_LAYERS_KEY, None)
        if n_layers is None:
            raise KeyError(f"Could not find n_layers in config file. Expected key {N_LAYERS_KEY}.")

        d_ff = config.get(D_FF_KEY, None)
        if d_ff is None:
            raise KeyError(f"Could not find d_ff in config file. Expected key {D_FF_KEY}.")

        vocab_size = config.get(VOCAB_SIZE_KEY) or len(self.tokenizer)

        model = Atacformer(
            vocab_size=vocab_size,
            d_model=d_model,
            n_heads=n_heads,
            n_layers=n_layers,
            d_ff=d_ff,
        )
        model.load_state_dict(params)

        self._model = model

        self.trained = True
        if POOLING_METHOD_KEY in config:
            self.pooling_method = config[POOLING_METHOD_KEY]

    def _init_from_huggingface(
        self,
        model_path: str,
        **kwargs,
    ):
        """
        Initialize the model from a huggingface model. This uses the model path
        to download the necessary files and then "build itself up" from those. This
        includes both the actual model and the tokenizer.

        :param str model_path: Path to the pre-trained model on huggingface.
        :param str model_file_name: Name of the model file.
        :param str universe_config_file_name: Name of the universe file.
        :param kwargs: Additional keyword arguments to pass to the hf download function.
        """
        model_file_name: str = MODEL_FILE_NAME
        universe_config_file_name: str = UNIVERSE_CONFIG_FILE_NAME
        universe_file_name: str = UNIVERSE_FILE_NAME
        config_file_name: str = CONFIG_FILE_NAME

        model_file_path = hf_hub_download(model_path, model_file_name, **kwargs)
        universe_config_path = hf_hub_download(model_path, universe_config_file_name, **kwargs)
        _universe_path = hf_hub_download(model_path, universe_file_name, **kwargs)
        config_path = hf_hub_download(model_path, config_file_name, **kwargs)

        self._load_local_model(model_file_path, universe_config_path, config_path)

    @classmethod
    def from_pretrained(cls, path_to_files: str) -> "AtacformerExModel":
        """
        Load the model from a set of files that were exported using the export function.

        :param str path_to_files: Path to the directory containing the files.
        :param str model_file_name: Name of the model file.
        :param str universe_config_file_name: Name of the universe file.
        """
        model_file_name: str = MODEL_FILE_NAME
        universe_config_file_name: str = UNIVERSE_CONFIG_FILE_NAME
        config_file_name: str = CONFIG_FILE_NAME

        model_file_path = os.path.join(path_to_files, model_file_name)
        universe_config_file_path = os.path.join(path_to_files, universe_config_file_name)
        config_file_path = os.path.join(path_to_files, config_file_name)

        instance = cls()
        instance._load_local_model(model_file_path, universe_config_file_path, config_file_path)
        instance.trained = True

        return instance

    def export(
        self,
        path: str,
        checkpoint_file: str = MODEL_FILE_NAME,
        _universe_config_file: str = UNIVERSE_CONFIG_FILE_NAME,
        config_file: str = CONFIG_FILE_NAME,
        **kwargs,
    ):
        """
        Function to facilitate exporting the model in a way that can
        be directly uploaded to huggingface. This exports the model
        weights and the vocabulary.

        :param str path: Path to export the model to.
        """
        # make sure the path exists
        if not os.path.exists(path):
            os.makedirs(path)

        # make sure the model is unwrapped
        if isinstance(self._model, nn.DataParallel):
            self._model = self._model.module
        elif isinstance(self._model, nn.parallel.DistributedDataParallel):
            self._model = self._model.module
        elif isinstance(self._model, nn.Module):
            pass
        else:
            raise TypeError("model must be an nn.Module object.")

        # export the model weights
        torch.save(self._model.state_dict(), os.path.join(path, checkpoint_file))

        d_model = self._model.d_model
        n_layers = self._model.n_layers
        n_heads = self._model.n_heads
        d_ff = self._model.d_ff
        vocab_size = len(self.tokenizer)
        context_size = self.context_size

        config = {
            POOLING_METHOD_KEY: self.pooling_method,
            D_MODEL_KEY: d_model,
            VOCAB_SIZE_KEY: vocab_size,
            N_HEADS_KEY: n_heads,
            N_LAYERS_KEY: n_layers,
            D_FF_KEY: d_ff,
            CONTEXT_SIZE_KEY: context_size,
        }

        if kwargs:
            config.update(kwargs)

        with open(os.path.join(path, config_file), "w") as f:
            safe_dump(config, f)

    def encode(self, adata: Union[str, sc.AnnData]) -> torch.Tensor:
        """
        Get the vector for a region.

        :param Region region: Region to get the vector for.
        :param str pooling: Pooling type to use.

        :return np.ndarray: Vector for the region.
        """
        raise NotImplementedError("This method is not implemented yet. Stay tuned...")



import lightning as L
import pytest
from torch.utils.data import DataLoader

from geniml.scembed.main import ScEmbed
from geniml.scembed.utils import BatchCorrectionDataset, BCBatchCollator

# from geniml.training.adapters import AdversarialBatchCorrectionAdapter


@pytest.fixture
def universe_file():
    return "tests/data/universe_mlm.bed"


@pytest.fixture
def data():
    return "tests/data/gtok_sample/"


@pytest.mark.skip("Not implemented")
class TestAtacFormer:
    def test_atacformer_dataset(self, data: str):
        dataset = BatchCorrectionDataset(
            [data, data]
        )  # just do the same thing twice, it doesnt matter this is just a test

        assert dataset is not None
        assert all([isinstance(x, tuple) for x in dataset])

    @pytest.mark.skip(reason="This test uses a pretrained model and is not suitable for CI")
    def test_adapter_init(
        self,
    ):
        model = ScEmbed("databio/r2v-luecken2021-hg38-v2")
        adapter = AdversarialBatchCorrectionAdapter(
            model=model, mode="adversary", num_batches=2, grad_rev_alpha=1.0
        )

        assert adapter is not None

    @pytest.mark.skip(reason="This test uses a pretrained model and is not suitable for CI")
    def test_train_with_adapter(self, universe_file: str, data: str):
        # get the model
        model = ScEmbed("databio/r2v-luecken2021-hg38-v2")

        # make adapter
        adapter = AdversarialBatchCorrectionAdapter(
            model=model, mode="adversary", num_batches=2, grad_rev_alpha=1.0
        )

        # make dataset
        collator = BCBatchCollator(model.tokenizer.padding_token_id())
        dataset = BatchCorrectionDataset(
            [data, data]
        )  # just do the same thing twice, it doesnt matter this is just a test

        # make dataloader
        dataloader = DataLoader(dataset, batch_size=2, shuffle=True, collate_fn=collator)

        # make trainer
        trainer = L.Trainer(max_epochs=3)

        # train in the adversarial mode
        trainer.fit(adapter, dataloader)

        # switch to the batch correction mode
        adapter.set_mode("batch_correction")

        # train in the batch correction mode
        trainer.fit(adapter, dataloader)



import gzip
import os
from hashlib import md5
from typing import Dict, List, Union

from .const import (
    MAF_CENTER_COL_NAME,
    MAF_CHROMOSOME_COL_NAME,
    MAF_COLUMN,
    MAF_END_COL_NAME,
    MAF_ENTREZ_GENE_ID_COL_NAME,
    MAF_FILE_DELIM,
    MAF_HUGO_SYMBOL_COL_NAME,
    MAF_NCBI_BUILD_COL_NAME,
    MAF_START_COL_NAME,
    MAF_STRAND_COL_NAME,
)


def is_gzipped(file: str) -> bool:
    """
    Check if a file is gzipped.

    :param file: path to file
    :return: True if file is gzipped, else False
    """
    _, file_extension = os.path.splitext(file)
    return file_extension == ".gz"


def extract_maf_col_positions(file: str) -> Dict[MAF_COLUMN, Union[int, None]]:
    """
    Extract the column positions of the MAF file.

    :param file: path to .maf file
    :return: dictionary of column positions
    """

    def get_index_from_header(header: List[str], col_name: str) -> Union[int, None]:
        """
        Get the index of a column from a header.

        :param header: list of column names
        :param col_name: column name
        :return: index of column
        """
        try:
            return header.index(col_name)
        except ValueError:
            return None

    # detect open function
    open_func = open if not is_gzipped(file) else gzip.open
    mode = "r" if not is_gzipped(file) else "rt"

    with open_func(file, mode) as f:
        header = f.readline().strip().split(MAF_FILE_DELIM)
        col_positions = {
            MAF_HUGO_SYMBOL_COL_NAME: get_index_from_header(header, MAF_HUGO_SYMBOL_COL_NAME),
            MAF_ENTREZ_GENE_ID_COL_NAME: get_index_from_header(
                header, MAF_ENTREZ_GENE_ID_COL_NAME
            ),
            MAF_CENTER_COL_NAME: get_index_from_header(header, MAF_CENTER_COL_NAME),
            MAF_NCBI_BUILD_COL_NAME: get_index_from_header(header, MAF_NCBI_BUILD_COL_NAME),
            MAF_CHROMOSOME_COL_NAME: get_index_from_header(header, MAF_CHROMOSOME_COL_NAME),
            MAF_START_COL_NAME: get_index_from_header(header, MAF_START_COL_NAME),
            MAF_END_COL_NAME: get_index_from_header(header, MAF_END_COL_NAME),
            MAF_STRAND_COL_NAME: get_index_from_header(header, MAF_STRAND_COL_NAME),
        }
    return col_positions


def read_bedset_file(file_path: str) -> List[str]:
    """
    Load a bedset from a text file

    :param file_path: path to the file

    :return: list of bed identifiers
    """
    bed_identifiers = []

    with open(file_path, "r") as f:
        for line in f:
            bed_identifiers.append(line.strip())
    return bed_identifiers


def compute_md5sum_bedset(bedset: List[str]) -> str:
    """
    Compute the md5sum of a bedset

    :param bedset: list of bed identifiers

    :return: md5sum of the bedset
    """
    return md5("".join(bedset).encode()).hexdigest()



class GenimlBaseError(Exception):
    """Base error type for peppy custom errors."""

    def __init__(self, msg):
        super(GenimlBaseError, self).__init__(msg)


class BBClientError(GenimlBaseError):
    """Base error type for BBClient errors."""

    def __init__(self, msg):
        super(BBClientError, self).__init__(msg)


class TokenizedFileNotFoundError(BBClientError):
    """Error raised when a tokenized file is not found."""

    def __init__(self, msg):
        super(TokenizedFileNotFoundError, self).__init__(msg)


class TokenizedFileNotFoundInCacheError(BBClientError):
    """Error raised when a tokenized file is not found in cache."""

    def __init__(self, msg):
        super(TokenizedFileNotFoundInCacheError, self).__init__(msg)



from typing import Literal

MODULE_NAME = "region2vec"
LR_TYPES = Literal["constant", "exponential", "step"]
POOLING_TYPES = Literal["mean", "max"]
MAX_WAIT_TIME = 10800

DEFAULT_EPOCHS = 100
DEFAULT_GENSIM_EPOCHS = 1
DEFAULT_MIN_COUNT = 10
DEFAULT_N_SHUFFLES = 1  # 1 is sufficient for most cases
DEFAULT_WINDOW_SIZE = 5
DEFAULT_EMBEDDING_DIM = 100
DEFAULT_EPOCHS = 10
DEFAULT_INIT_LR = 0.1  # https://github.com/databio/gitk/issues/6#issuecomment-1476273162
DEFAULT_MIN_LR = 0.0001  # gensim default
DEFAULT_NS_POWER = 0.75

CONFIG_FILE_NAME = "config.yaml"
MODEL_FILE_NAME = "checkpoint.pt"
UNIVERSE_FILE_NAME = "universe.bed"

POOLING_METHOD_KEY = "pooling_method"
EMBEDDING_DIM_KEY = "embedding_dim"
EMBEDDING_DIM_KEY_OLD = "embedding_size"
VOCAB_SIZE_KEY = "vocab_size"



import random
from typing import Dict, List, Tuple, Union

import numpy as np


def verify_load_inputs(
    vectors: np.ndarray,
    ids: Union[List[Union[str, int]], np.ndarray],
    payloads: List[Dict[str, str]],
):
    n_ids = len(ids)
    n_vectors = vectors.shape[0]
    n_payloads = len(payloads)
    if n_ids != n_vectors or n_ids != n_payloads:
        raise ValueError(
            "The number of ids ({n_ids}), vectors ({n_vectors}), and payloads ({n_payloads}) must match"
        )


def single_query_eval(search_results: List, relevant_results: List) -> Tuple[float, float, float]:
    """
    Evaluate a single query

    :param search_results: List of store ids, by the order of similarity in search
    :param relevant_results: List if store id which are relevant search results

    :return: a Tuple of (Average Precision, AUC-ROC, R-precision)
    """
    num_relevant = len(relevant_results)
    retrieved_relevant = 0
    k = len(search_results)
    sum_precision = 0
    x = [0]  # (false_positive/(false_positive + true_negative)
    y = [0]  # recall or  true_positive / (true_positive + false_negative)
    false_positive = 0
    true_negative = k - num_relevant
    true_positive = 0
    false_negative = num_relevant

    for i in range(k):
        result = search_results[i]
        result_id = result["id"]
        if result_id in relevant_results:  # one relevant is retrieved
            true_positive += 1
            false_negative -= 1
            retrieved_relevant += 1

            sum_precision += retrieved_relevant / (i + 1)

        else:  # one irrelevant is retrieved
            false_positive += 1
            true_negative -= 1
        x.append(false_positive / (false_positive + true_negative))
        y.append(true_positive / (true_positive + false_negative))
        if i == num_relevant - 1:
            r_precision = retrieved_relevant / num_relevant
    average_precision = sum_precision / num_relevant
    # compute AUC-ROC
    auc = np.trapz(y, x)
    return average_precision, auc, r_precision


def rand_eval(n: int, query_dict: Dict) -> Tuple[float, float, float]:
    """
    Evaluation results if the retrieval is completely random

    :param n: total number of results

    :param query_dict:  a dictionary that contains query and relevant results in this format:
        {
            <query string>:[
                <store id in backend>,
                ...
            ],
            ...
        }

    :return: a Tuple of (Average Precision, AUC-ROC, R-precision)
    """
    sum_ap = 0  # sum of all average precisions
    sum_auc = 0
    sum_rp = 0
    query_count = 0

    for query_str in query_dict.keys():
        relevant_results = query_dict[query_str]  # set of relevant ids
        search_results_ids = list(range(n))
        random.shuffle(search_results_ids)
        search_results = []
        for result_id in search_results_ids:
            search_results.append({"id": result_id})
        query_count += 1
        ap, auc, rp = single_query_eval(search_results, relevant_results)
        sum_ap += ap
        sum_auc += auc
        sum_rp += rp

    if query_count > 0:
        return sum_ap / query_count, sum_auc / query_count, sum_rp / query_count

    else:
        return 0.0, 0.0, 0.0



import os
from functools import cmp_to_key
from logging import getLogger

import numpy as np
import pyBigWig
from scipy.stats import nbinom

from ..const import PKG_NAME
from ..utils import natural_chr_sort
from .const import LAMBDAS, TRANSMAT
from .models import PoissonModel
from .utils import find_full, predictions_to_bed

_LOGGER = getLogger(PKG_NAME)

""" States legend
0 -> start
1 -> core
2 -> end
3 -> background"""


def norm(track, mode):
    """Normalize the coverage track depending on track type.
    For each unique value in the track calculates the corresponding
    quantile taking into account that values occur different number of times."""
    important_val = track[track != 0]
    important_val_unique, counts = np.unique(important_val, return_counts=True)
    uniq_dict = {i: j for i, j in zip(important_val_unique, counts)}
    # how many times each value is present in the track
    important_val_unique_sort = np.sort(important_val_unique)
    if mode == "ends":
        n = 0.1
    if mode == "core":
        n = 0.2
    bs = 0  # what fraction of the distribution was used for normalization
    val = {}  # for each unique value in track holds the corresponding quantile
    for i in important_val_unique_sort:
        move_val = (uniq_dict[i] / len(important_val)) / 2
        # how far from last quantile is te next one
        val[i] = nbinom.ppf(bs + move_val, 1, n)
        bs = bs + move_val * 2
    track[track != 0] = [val[i] for i in important_val]


def process_bigwig(file, seq, p, chrom, chrom_size, normalize=True, mode=None):
    """Preprocess bigWig file"""
    if pyBigWig.numpy:
        track = file.values(chrom, 0, chrom_size, numpy=True)
    else:
        track = file.values(chrom, 0, chrom_size)
        track = np.array(track)
    track[np.isnan(track)] = 0
    track = track.astype(np.uint16)
    if normalize:
        norm(track, mode)
    seq[:, p] = track


def read_data(start, core, end, chrom, normalize=True):
    """
    Read in and preprocess data
    :param str start: path to file with start coverage
    :param str end: path to file with end coverage
    :param str core: path to file with  core coverage
    :param str chrom: chromosome to analyse
    :param bool normalize: whether to normalize the coverage
    :return: chromosome size, coverage matrix
    """
    start = pyBigWig.open(start + ".bw")
    chroms = start.chroms()
    chrom_size = chroms[chrom]
    seq = np.zeros((chrom_size, 3), dtype=np.uint16)
    process_bigwig(start, seq, 0, chrom, chrom_size, normalize, mode="ends")
    start.close()
    core = pyBigWig.open(core + ".bw")
    process_bigwig(core, seq, 1, chrom, chrom_size, normalize, mode="core")
    core.close()
    end = pyBigWig.open(end + ".bw")
    process_bigwig(end, seq, 2, chrom, chrom_size, normalize, mode="ends")
    end.close()
    return chrom_size, seq


def split_predict(seq, empty_starts, empty_ends, model):
    """Make model prediction only for regions containing
    nonzero positions"""
    hmm_predictions = np.full(len(seq), 3, dtype=np.uint8)
    for s, e in zip(empty_starts, empty_ends):
        res = model.predict(seq[s:e])
        hmm_predictions[s:e] = res
    return hmm_predictions


def run_hmm(start, core, end, chrom, normalize=True):
    """Make HMM prediction for given chromosome"""
    chrom_size, seq = read_data(start, core, end, chrom, normalize=normalize)
    empty_starts, empty_ends = find_full(seq)
    model = PoissonModel(TRANSMAT, LAMBDAS, save_matrix=False).model
    hmm_predictions = split_predict(seq, empty_starts, empty_ends, model)
    return hmm_predictions, model


def hmm_universe(
    coverage_folder,
    out_file,
    prefix="all",
    normalize=True,
    save_max_cove=False,
):
    """
    Create HMM based universe from coverage
    :param str coverage_folder: path to folder with coverage files
    :param str start: start coverage file name
    :param str end: end coverage file name
    :param str core: core coverage file name
    :param str out_file: path to the output file with universe
    :param bool normalize: whether to normalize file
    :param bool save_max_cove: whether to save the maximum
    peak coverage
    """
    if os.path.isfile(out_file):
        raise Exception(f"File : {out_file} exists")
    start = os.path.join(coverage_folder, f"{prefix}_start")
    core = os.path.join(coverage_folder, f"{prefix}_core")
    end = os.path.join(coverage_folder, f"{prefix}_end")
    bw_start = pyBigWig.open(start + ".bw")
    chroms = bw_start.chroms()
    bw_start.close()
    chroms_key = list(chroms.keys())
    chroms_key = sorted(chroms_key, key=cmp_to_key(natural_chr_sort))
    chroms = {i: chroms[i] for i in chroms_key}
    for C in chroms:
        if chroms[C] > 0:
            pred, m = run_hmm(start, core, end, C, normalize=normalize)
            predictions_to_bed(
                pred,
                C,
                out_file,
                save_max_cove=save_max_cove,
                cove_file=core + ".bw",
            )


def test_hmm(message):
    """Just prints a test message"""
    _LOGGER.info(message)



import torch
import torch.nn as nn
import torch.nn.functional as F


class Attention(nn.Module):
    def __init__(self, embed_dim):
        super(Attention, self).__init__()
        self.embed_dim = embed_dim
        self.attention = nn.Linear(embed_dim, 1)

    def forward(self, x):
        # x: [batch_size, seq_len, embed_dim]
        attention_weights = F.softmax(self.attention(x), dim=1)
        # attention_weights: [batch_size, seq_len, 1]
        weighted_sum = torch.sum(x * attention_weights, dim=1) / torch.sum(
            attention_weights, dim=1
        )
        # weighted_sum: [batch_size, embed_dim]
        return weighted_sum


# https://github.com/tadeephuy/GradientReversal
class GradientReversalLayer(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, alpha):
        ctx.alpha = alpha
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        output = grad_output.neg() * ctx.alpha
        return output, None


class GradientReversal(nn.Module):
    def __init__(self, alpha):
        super().__init__()
        self.alpha = torch.tensor(alpha, requires_grad=False)

    def forward(self, x):
        return GradientReversalLayer.apply(x, self.alpha)



from qdrant_client.http import models
from qdrant_client.models import Distance

DEFAULT_QDRANT_HOST = "localhost"
DEFAULT_QDRANT_PORT = 6333

DEFAULT_COLLECTION_NAME = "embeddings"

DEFAULT_QDRANT_DIST = Distance.COSINE

DEFAULT_INDEX_PATH = "./current_index.bin"

DEFAULT_HNSW_SPACE = "cosine"

DEFAULT_DIM = 100

DEFAULT_TEXT_DIM = 384


# the size of the dynamic list for the nearest neighbors
# Higher ef leads to more accurate but slower search
# cannot be set lower than the number of queried nearest neighbors k
DEFAULT_EF = 200

# the number of bi-directional links created for every new element during construction
# Higher M work better on datasets with high intrinsic dimensionality and/or high recall
# low M work better for datasets with low intrinsic dimensionality and/or low recalls.
DEFAULT_M = 64

DEFAULT_QUANTIZATION_CONFIG = models.ScalarQuantization(
    scalar=models.ScalarQuantizationConfig(
        type=models.ScalarType.INT8,
        quantile=0.99,
        always_ram=True,
    ),
)


# for evaluation dataset from huggingface
HF_INDEX = "index.bin"
HF_PAYLOADS = "payloads.pkl"
HF_METADATA = "metadata.json"

TEXT_ENCODER_REPO = "databio/bivec-search-demo"



import os

import genomicranges
import pandas as pd
import pytest

from geniml.io.exceptions import BEDFileReadError, GenimlBaseError
from geniml.io.io import SNP, Maf, Region, RegionSet

DATA_TEST_FOLDER = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "tests",
    "data",
    "io_data",
)
DATA_TEST_FOLDER_BED = os.path.join(DATA_TEST_FOLDER, "bed")
DATA_TEST_FOLDER_MAF = os.path.join(DATA_TEST_FOLDER, "maf")
DATA_TEST_FOLDER_BED_BAD = os.path.join(DATA_TEST_FOLDER, "bed_bad")

ALL_BEDFILE_PATH = [
    os.path.join(DATA_TEST_FOLDER_BED, x) for x in os.listdir(DATA_TEST_FOLDER_BED)
]
ALL_MAF_PATH = [os.path.join(DATA_TEST_FOLDER_MAF, x) for x in os.listdir(DATA_TEST_FOLDER_MAF)]
ALL_BADFILE_BAD_PATH = [
    os.path.join(DATA_TEST_FOLDER_BED_BAD, x) for x in os.listdir(DATA_TEST_FOLDER_BED_BAD)
]


def test_make_region():
    r = Region("chr1", 0, 100)
    assert r is not None
    assert r.chr == "chr1"
    assert r.start == 0
    assert r.end == 100


def test_make_SNP():
    s = SNP(
        hugo_symbol="TP53",
        chromosome="chr1",
        start_position=0,
        end_position=1,
        ncbi_build="GRCh38",
        strand="+",
    )
    assert s is not None
    assert s.hugo_symbol == "TP53"
    assert s.chromosome == "chr1" and s.chr == "chr1"
    assert s.start_position == 0 and s.start == 0
    assert s.end_position == 1 and s.end == 1
    assert s.ncbi_build == "GRCh38"
    assert s.strand == "+"


class TestRegionSet:
    @pytest.mark.parametrize(
        "url",
        [
            "https://github.com/databio/geniml/raw/master/tests/data/io_data/bed/s1_a.bed.gz",
        ],
    )
    def test_region_set_from_url(self, url):
        region_set = RegionSet(url)
        assert isinstance(region_set.regions[0], Region)

    @pytest.mark.parametrize("url", ALL_BEDFILE_PATH)
    def test_region_set_from_path_backed(self, url):
        region_set = RegionSet(url)
        assert isinstance(region_set.regions[0], Region)

    @pytest.mark.parametrize("url", ALL_BEDFILE_PATH)
    def test_region_set_from_path(self, url):
        region_set = RegionSet(url, backed=True)
        for region in region_set:
            assert isinstance(region, Region)
            break

    @pytest.mark.parametrize("path", ALL_BADFILE_BAD_PATH)
    def test_broken_bed_from_path(self, path):
        with pytest.raises(BEDFileReadError):
            region_set = RegionSet(path)

    @pytest.mark.parametrize(
        "url",
        ["https://github.com/databio/geniml/raw/master/tests/data/io_data/bed/s1_a.bed.gz"],
        # TODO: This is not the right way how to do it!
    )
    def test_region_set_from_url_cant_be_backed(self, url):
        with pytest.raises(GenimlBaseError):
            RegionSet(url, backed=True)

    def test_make_region_set_with_list(
        self,
    ):
        regions = [Region("chr1", 0, 100), Region("chr1", 100, 200)]
        u = RegionSet(regions)
        assert u is not None
        assert len(u) == 2

        # test we can iterate over it
        for region in u:
            assert isinstance(region, Region)

    @pytest.mark.parametrize("path", ALL_BEDFILE_PATH)
    def test_to_genomic_ranges(self, path):
        bedfile = RegionSet(path)
        gr = bedfile.to_granges()  # should return a GenomicRanges object
        assert isinstance(gr, genomicranges.GenomicRanges)

    def test_calculation_id(self):
        bedfile_id_1 = RegionSet(ALL_BEDFILE_PATH[0]).identifier
        bedfile_id_2 = RegionSet(ALL_BEDFILE_PATH[1]).identifier
        bedfile_id_3 = RegionSet(ALL_BEDFILE_PATH[2]).identifier
        assert len(bedfile_id_2) == 32
        assert bedfile_id_1 == bedfile_id_2 == bedfile_id_3

    @pytest.mark.parametrize("url", ALL_BEDFILE_PATH)
    def test_to_df(self, url):
        region_set = RegionSet(url, backed=False)
        pandas_df = region_set.to_pandas()
        assert isinstance(pandas_df, pd.DataFrame)

    @pytest.mark.parametrize("url", ALL_BEDFILE_PATH)
    def test_to_df_backed(self, url):
        region_set = RegionSet(url, backed=True)
        pandas_df = region_set.to_pandas()
        assert isinstance(pandas_df, pd.DataFrame)


class TestMaff:
    @pytest.mark.parametrize("path", ALL_MAF_PATH)
    def test_maf_from_path(self, path):
        snps = Maf(path)
        assert snps is not None
        for s in snps:
            assert isinstance(s, SNP)
            break

    @pytest.mark.parametrize("path", ALL_MAF_PATH)
    def test_read_maf_file_backed(
        self,
        path: str,
    ):
        snps = Maf(path, backed=True)
        assert snps is not None
        assert len(snps) == 99
        for s in snps:
            assert isinstance(s, SNP)

    @pytest.mark.parametrize("path", ALL_MAF_PATH)
    def test_snps_to_regions(
        self,
        path: str,
    ):
        snps = Maf(path)
        for snp in snps:
            assert isinstance(snp, SNP)
            assert isinstance(snp.to_region(), Region)



import multiprocessing
import os
from typing import List

import numpy as np

from . import utils
from .region2vec_train import main as region2_train
from .region_shuffling import main as sent_gen


class Namespace:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def region2vec(
    token_folder: str,
    save_dir: str,
    file_list: List[str] = None,
    data_type: str = "files",
    mat_path: str = None,
    num_shufflings: int = 1000,
    num_processes: int = 10,
    tokenization_mode: str = "hard",
    embedding_dim: int = 100,
    context_win_size: int = 5,
    save_freq: int = -1,
    resume_path: str = "",
    train_alg: str = "cbow",
    min_count: int = 5,
    neg_samples: int = 5,
    init_lr: float = 0.025,
    min_lr: float = 1e-4,
    lr_scheduler: str = "linear",
    milestones: List[int] = [],
    hier_softmax: bool = False,
    seed: int = 0,
    update_vocab: str = "once",
):
    """Trains a Region2Vec model.

    Starts two subprocesses: one that generates shuffled datasets, and the
    other consumes the shuffled datasets to train a Region2Vec model.

    Args:
        token_folder (str): The path to the folder of tokenized files.
        save_dir (str): The folder that stores the training results.
        file_list (list[str], optional): Specifies which files from
            token_folder are used for training. When None, uses all the files
            in token_folder. Defaults to None.
        data_type (str, optional): "files" or "matrix". Defaults to "files".
        mat_path (str, optional): Used only when data_type = "matrix". Defaults
            to None.
        num_shufflings (int, optional): Number of shuffled datasets to
            generate. Defaults to 1000.
        num_processes (int, optional): Number of processes used. Defaults to 10.
        tokenization_mode (str, optional): Tokenization mode. Defaults to
            "hard", i.e., concatenating all regions in a BED files in a random order.
        embedding_dim (int, optional): Dimension of embedding vectors. Defaults
            to 100.
        context_win_size (int, optional): Context window size. Defaults to 5.
        save_freq (int, optional): Save frequency. Defaults to -1.
        resume_path (str, optional): Starts with a previously trained model.
            Defaults to "".
        train_alg (str, optional): Training algorithm. Defaults to "cbow".
        min_count (int, optional): Minimum frequency required to keep a region.
            Defaults to 5.
        neg_samples (int, optional): Number of negative samples used in
            training. Defaults to 5.
        init_lr (float, optional): Initial learning rate. Defaults to 0.025.
        min_lr (float, optional): Minimum learning rate. Defaults to 1e-4.
        lr_scheduler (str, optional): Type of the learning rate scheduler.
            Defaults to "linear".
        milestones (list[int], optional): Used only when
            lr_scheduler="milestones". Defaults to [].
        hier_softmax (bool, optional): Whether to use hierarchical softmax
            during training. Defaults to False.
        seed (int, optional): Random seed. Defaults to 0.
        update_vocab (str, optional): If "every", then updates the vocabulary
            for each shuffled dataset. Defaults to "once" assuming no new
            regions occur in shuffled datasets.
    """
    timer = utils.Timer()
    start_time = timer.t()
    if file_list is None:
        files = os.listdir(token_folder)
    else:
        files = file_list
    os.makedirs(save_dir, exist_ok=True)
    file_list_path = os.path.join(save_dir, "file_list.txt")
    utils.set_log_path(save_dir)
    with open(file_list_path, "w") as f:
        for file in files:
            f.write(file)
            f.write("\n")

    training_processes = []
    num_sent_processes = min(int(np.ceil(num_processes / 2)), 4)
    nworkers = min(num_shufflings, num_sent_processes)
    utils.log(f"num_sent_processes: {nworkers}")
    if nworkers <= 1:
        sent_gen_args = Namespace(
            tokenization_folder=token_folder,
            save_dir=save_dir,
            file_list=file_list_path,
            tokenization_mode=tokenization_mode,
            pool=1,  # maximum number of unused shuffled datasets generated at a time
            worker_id=0,
            number=num_shufflings,
        )
        p = multiprocessing.Process(target=sent_gen, args=(sent_gen_args,))
        p.start()
        training_processes.append(p)
    else:
        num_arrs = [num_shufflings // nworkers] * (nworkers - 1)

        num_arrs.append(num_shufflings - np.array(num_arrs).sum())
        sent_gen_args_arr = []
        for n in range(nworkers):
            sent_gen_args = Namespace(
                tokenization_folder=token_folder,
                data_type=data_type,
                mat_path=mat_path,
                save_dir=save_dir,
                file_list=file_list_path,
                tokenization_mode=tokenization_mode,
                pool=1,  # maximum number of unused shuffled datasets generated at a time
                worker_id=n,
                number=num_arrs[n],
            )
            sent_gen_args_arr.append(sent_gen_args)
        for n in range(nworkers):
            p = multiprocessing.Process(target=sent_gen, args=(sent_gen_args_arr[n],))
            p.start()
            training_processes.append(p)

    num_region2vec_processes = max(num_processes - nworkers, 1)
    region2vec_args = Namespace(
        num_shuffle=num_shufflings,
        embed_dim=embedding_dim,
        context_len=context_win_size,
        nworkers=num_region2vec_processes,
        save_freq=save_freq,
        save_dir=save_dir,
        resume=resume_path,
        train_alg=train_alg,
        min_count=min_count,
        neg_samples=neg_samples,
        init_lr=init_lr,
        min_lr=min_lr,
        lr_mode=lr_scheduler,
        milestones=milestones,
        hier_softmax=hier_softmax,
        update_vocab=update_vocab,
        seed=seed,
    )
    p = multiprocessing.Process(target=region2_train, args=(region2vec_args,))
    p.start()
    training_processes.append(p)
    for p in training_processes:
        p.join()
    os.remove(file_list_path)
    elapsed_time = timer.t() - start_time
    print(f"[Training] {utils.time_str(elapsed_time)}/{utils.time_str(timer.t())}")



import argparse
import datetime
import glob
import logging
import os
import pickle
import random
import time
from typing import Union

from gensim.models import Word2Vec
from gensim.models.word2vec import LineSentence

from . import utils
from .const import *


def find_dataset(data_folder: str) -> Union[str, int]:
    """Finds an available dataset in data_folder.

    Finds an available dataset from a folder specified by data_folder.

    Args:
        data_folder (str): The path to the folder where datasets are generated.

    Returns:
        Union[str, int]: -1 when no datasets are found after MAX_WAIT_TIME;
            otherwise, a path to the found dataset.
    """
    train_pattern = os.path.join(data_folder, "pool*[0-9]")
    count = 0
    while True:
        dsets = glob.glob(train_pattern)
        if len(dsets) == 0:
            print("No available dataset, waiting for generation...", end="\r")
            time.sleep(1)
            count += 1
            if count == MAX_WAIT_TIME:
                print("Wait time exceeds MAX_WAIT_TIME, exit")
                return -1
        else:
            return dsets[random.randint(0, len(dsets) - 1)]


def main(args: argparse.Namespace) -> None:
    """Trains a Region2Vec model using the arguments in args.

    Called internally as a subprocess by region2vec in main.py.

    Args:
        args (argparse.Namespace): See the definition of ArgumentParser.
    """
    save_dir = args.save_dir
    data_folder = os.path.join(
        save_dir, "shuffled_datasets"
    )  # shuffled datasets are stored in the shuffled_datasets folder
    model_dir = os.path.join(save_dir, "models")  # model snapshots are stored in the models folder
    os.makedirs(save_dir, exist_ok=True)
    os.makedirs(model_dir, exist_ok=True)
    utils.set_log_path(save_dir)  # specify the path to the log file
    utils.log(str(vars(args)))
    # ----------- configure and prepare the model -----------
    if args.train_alg == "skip-gram":
        train_alg = 1
        msg_model = "\033[94mUsing skip-gram, "
    else:
        train_alg = 0
        msg_model = "\033[94mUsing cbow, "
    if args.hier_softmax == True or args.neg_samples == 0:
        hs = 1
        msg_model += "hierarchical softmax\033[00m"
    else:
        hs = 0
        msg_model += f"negative sampling with {args.neg_samples} negative samples\033[00m"
    if not os.path.exists(args.resume):
        vocab_update = False
        model = Word2Vec(
            vector_size=args.embed_dim,
            alpha=args.init_lr,
            window=args.context_len,
            min_count=args.min_count,
            seed=args.seed,
            workers=args.nworkers,
            sg=train_alg,
            negative=args.neg_samples,
            hs=hs,
        )
        utils.log(msg_model)
    else:
        utils.log(
            f"\033[91mResuming {args.resume}, make sure the model configurations are consistent\033[00m"
        )
        model = Word2Vec.load(args.resume)
        vocab_update = True

    # ----------- train the model -----------
    loss_all = []
    if args.lr_mode == "milestone":
        lr_info = {"milestones": args.milestones, "ratio": 0.1}
    elif args.lr_mode == "linear":
        lr_info = {"freq": 1}

    lr_scheduler = utils.lr_scheduler(
        args.init_lr,
        args.min_lr,
        args.num_shuffle,
        lr_info=lr_info,
        mode=args.lr_mode,
    )

    run_timer = utils.Timer()
    cur_time = datetime.datetime.now().strftime("%x-%X")
    utils.log(f"[{cur_time}] Start training")
    utils.log(f"[{cur_time}] Building vocabulary")
    dset = find_dataset(data_folder)
    if dset == -1:
        return
    sentences = LineSentence(dset)  # create sentence iterator
    model.build_vocab(sentences, update=vocab_update)  # prepare the model vocabulary
    cur_time = datetime.datetime.now().strftime("%x-%X")
    utils.log(f"[{cur_time}] Vocabulary size is {len(model.wv.index_to_key)}")
    build_vocab_time = run_timer.t()
    min_loss = 1.0e100

    # start training
    for sidx in range(args.num_shuffle):
        epoch_timer = utils.Timer()
        msg = f"[Shuffling {sidx + 1:>4d}] "
        dset = find_dataset(data_folder)
        if dset == -1:
            return
        dname = dset.split("/")[-1]
        dst_name = os.path.join(data_folder, dname + "using")
        os.rename(dset, dst_name)  # change to file name to pool%dusing
        sentences = LineSentence(dst_name)  # create sentence iterator
        if args.update_vocab == "every":
            model.build_vocab(sentences, update=True)  # prepare the model vocabulary
        model.train(
            sentences,
            total_examples=model.corpus_count,
            epochs=1,
            compute_loss=True,
            start_alpha=lr_scheduler.lr,
            end_alpha=lr_scheduler.lr,
        )

        loss = model.get_latest_training_loss()
        loss_all.append(loss)
        used_name = os.path.join(data_folder, dname + "used")
        os.rename(dst_name, used_name)

        if loss < min_loss:
            min_loss = loss
            model.save(os.path.join(model_dir, "region2vec_best.pt"))
        model.save(os.path.join(model_dir, "region2vec_latest.pt"))
        if args.save_freq > 0 and (sidx + 1) % args.save_freq == 0:
            model.save(os.path.join(model_dir, f"region2vec_{sidx + 1}.pt"))
        est_time = (run_timer.t() - build_vocab_time) / (
            sidx + 1
        ) * args.num_shuffle + build_vocab_time
        msg += f"loss {loss:>12.4f} lr {lr_scheduler.lr:>5.4f} vocab_size {len(model.wv.index_to_key):>12d} ({utils.time_str(epoch_timer.t())}/{utils.time_str(est_time)})"
        utils.log(msg)
        lr_scheduler.step()

    with open(os.path.join(model_dir, "loss_all.pickle"), "wb") as f:
        pickle.dump(loss_all, f)

    elapsed_time = run_timer.t()
    cur_time = datetime.datetime.now().strftime("%x-%X")
    utils.log(f"[{cur_time}] Training finished, training Time {utils.time_str(elapsed_time)}")
    # remove intermediate datasets
    os.system(f"rm -rf {data_folder}")  # remove the generated shuffled datasets


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Region2Vec training")
    parser.add_argument(
        "--num-shuffle", type=int, help="number of shuffled datasets used for training"
    )
    parser.add_argument("--embed-dim", type=int, help="dimension of embedding vectors")
    parser.add_argument("--context-len", type=int, help="length of the moving context window")
    parser.add_argument(
        "--nworkers", type=int, help="number of parallel processes used in training"
    )
    parser.add_argument(
        "--save-freq",
        type=int,
        default=0,
        help="save the Region2Vec model after consuming save_freq datasets; use a nonpositive number to disable",
    )
    parser.add_argument("--save-dir", help="the folder that saves the Region2Vec model")
    parser.add_argument("--resume", help="path to a previously trained model")
    parser.add_argument("--train-alg", help="training algorithm, select from [cbow, skip-gram]")
    parser.add_argument(
        "--min-count",
        type=int,
        help="threshold of pruning the internal vocabulary",
    )
    parser.add_argument(
        "--neg-samples",
        type=int,
        help="number of noise words in negative sampling, usually between 5-20",
    )
    parser.add_argument(
        "--hier-softmax",
        default=False,
        action="store_true",
        help="if given, hierarchical softmax will be used",
    )
    parser.add_argument("--init-lr", type=float, help="initial learning rate")
    parser.add_argument("--milestones", nargs="+", type=int, default=[100, 200])
    parser.add_argument(
        "--lr-mode",
        type=str,
        choices=["milestone", "linear"],
        help="type of learning rate scheduler, milestone or linear",
    )
    parser.add_argument(
        "--update-vocab",
        type=str,
        default="once",
        help="[every] update at every epoch; [once] Update once since the vocabulary does not change",
    )
    parser.add_argument("--min-lr", type=float, help="minimum learning rate")
    parser.add_argument("--seed", type=int, help="random seed")
    args = parser.parse_args()
    main(args)



from abc import ABC, abstractmethod
from typing import Union

import numpy as np

from ...io import RegionSet


class Query2Vec(ABC):
    """
    An abstract class representing query embedder. In retrieval of region sets,
    it embeds the query into a vector, which is used for KNN search in backend
    """

    def __init__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def forward(self, query: Union[str, RegionSet]) -> np.ndarray:
        """

        Parameters
        ----------
        query : a natural language string (query term or path to a BED file in disk),
        or a RegionSet object

        Returns
        -------
        the embedding vector
        """
        raise NotImplementedError



import logging
import os
from glob import glob
from typing import List, Tuple

import scanpy as sc
import torch
from gtars.utils import read_tokens_from_gtok
from rich.progress import track
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from .const import DEFAULT_CHUNK_SIZE

_LOGGER = logging.getLogger(__name__)


class AnnDataChunker:
    def __init__(self, adata: sc.AnnData, chunk_size: int = DEFAULT_CHUNK_SIZE):
        """
        Simple class to chunk an AnnData object into smaller pieces. Useful for
        training on large datasets.

        :param sc.AnnData adata: AnnData object to chunk. Must be in backed mode. See: https://scanpy.readthedocs.io/en/stable/generated/scanpy.read_h5ad.html
        :param int chunk_size: Number of cells to include in each chunk
        """
        self.adata = adata
        self.chunk_size = chunk_size
        self.n_chunks = len(adata) // chunk_size + 1

    def __iter__(self):
        for i in range(self.n_chunks):
            # check for shape = 0
            if self.adata[i * self.chunk_size : (i + 1) * self.chunk_size, :].shape[0] == 0:
                return
            yield self.adata[i * self.chunk_size : (i + 1) * self.chunk_size, :]

    def __len__(self):
        return self.n_chunks

    def __getitem__(self, item: int):
        """
        Get a chunk of the AnnData object.

        :param int item: The chunk index to get.
        """
        return self.adata[item * self.chunk_size : (item + 1) * self.chunk_size, :]

    def __repr__(self):
        return f"<AnnDataChunker: {self.n_chunks} chunks of size {self.chunk_size}>"


class BatchCorrectionDataset(Dataset):
    def __init__(self, batches: list, backed: bool = True):
        """
        Dataset for batch correction. This dataset takes in pre-tokenized
        cells and their batch of origin and then yields them out for training.

        :param batches list: a list of paths that point to pre-tokenized cells (.gtok files).
        :param bool backed: Whether to load the data in backed mode. If True, the data will be loaded in backed mode. If False, the data will be loaded into memory.
        """
        self.num_batches = len(batches)
        self.backed = backed

        # create tuples of (gtok_file, batch)
        self.data = []
        for i, batch in enumerate(batches):
            for gtok_file in glob(os.path.join(batch, "*.gtok")):
                self.data.append((gtok_file, i))

        if not self.backed:
            _LOGGER.info("Loading data into memory...")
            self.data = [
                (torch.tensor(read_tokens_from_gtok(gtok_file)), torch.tensor(batch))
                for gtok_file, batch in track(
                    self.data,
                    total=len(self.data),
                    description="Loading data into memory...",
                )
            ]

    def __getitem__(self, idx) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Get a single item from the dataset.

        :param idx: The index of the item to get.
        """
        if not self.backed:
            return self.data[idx]
        else:
            gtok_file, batch = self.data[idx]
            tokens = read_tokens_from_gtok(gtok_file)
            return torch.tensor(tokens), torch.tensor(batch)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return (
            f"<BatchCorrectionDataset: {len(self.data)} samples, and {self.num_batches} batches>"
        )


class BCBatchCollator:
    def __init__(self, pad_value: int = 0):
        self.pad_value = pad_value

    def __call__(
        self, batch: List[Tuple[torch.Tensor, torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Collate function for the batch correction dataset. This function
        takes in a list of tuples of (tokens, batch) and returns a tuple of
        (padded_tokens, batches).

        :param batch: A list of tuples of (tokens, batch)
        """
        tokens, batches = zip(*batch)
        tokens = pad_sequence(tokens, batch_first=True, padding_value=self.pad_value)
        batches = torch.stack(batches)
        return tokens, batches



import logging
from typing import Dict, List, Union

import numpy as np

from geniml.const import PKG_NAME

from ..backends import HNSWBackend, QdrantBackend
from ..query2vec import Text2Vec
from ..utils import single_query_eval
from .abstract import BEDSearchInterface

_LOGGER = logging.getLogger(PKG_NAME)


class Text2BEDSearchInterface(BEDSearchInterface):
    """Search interface for the query that is a natural language string"""

    def __init__(
        self,
        backend: Union[QdrantBackend, HNSWBackend],
        query2vec: Text2Vec,
    ):
        """
        Initiate the search interface

        :param backend: the backend where vectors are stored
        :param query2vec: a Text2Vec, for details, see docstrings in geniml.search.query2vec.text2vec
        """
        self.query2vec = query2vec
        self.backend = backend

    def query_search(
        self,
        query: Union[str, np.ndarray],
        limit: int,
        with_payload: bool = True,
        with_vectors: bool = True,
        offset: int = 0,
    ) -> List[Dict]:
        """

        :param query: the natural language query string, or a vector in the embedding space of region sets

        for rest of the parameters, check the docstring of QdrantBackend.search() or HNSWBackend.search()
        """
        if isinstance(query, np.ndarray):
            search_vec = query
        else:
            search_vec = self.query2vec.forward(query)

        return self.backend.search(search_vec, limit, with_payload, with_vectors, offset)

    def eval(self, query_dict: Dict[str, List[Union[int, np.int64]]]) -> Dict[str, float]:
        """
        With a query dictionary, return the Mean Average Precision, AUC-ROC and R-precision of query retrieval

        :param query_dict:a dictionary that contains query and relevant results in this format:
            {
                <query string>:[
                    <store id in backend>,
                    ...
                ],
                ...
            }

        :return: a Tuple of (Mean Average Precision, Average AUC-ROC, Average R-precision)
        """

        # number
        n = len(self.backend)

        # set ef for search
        # ef cannot be set lower than the number of queried nearest neighbors k

        if isinstance(self.backend, HNSWBackend):
            self.backend.idx.set_ef(n)

        sum_ap = 0  # sum of all average precision
        sum_auc = 0  # sum of all AUC-ROC
        sum_rp = 0  # sum of all R-Precision

        # total number of queries
        query_count = 0

        k = n  # to rank all results
        # evaluate each retrieval
        for query_str in query_dict.keys():
            relevant_results = query_dict[query_str]  # set of relevant ids

            try:
                search_results = self.query_search(
                    query=query_str, limit=k, with_vectors=False, with_payload=False
                )
                query_count += 1
            except Exception as e:
                _LOGGER.error(f"{query_str} caused {e}")
                continue

            ap, auc, rp = single_query_eval(search_results, relevant_results)
            sum_ap += ap
            sum_auc += auc
            sum_rp += rp

        if query_count > 0:
            return {
                "Mean Average Precision": sum_ap / query_count,
                "Mean AUC-ROC": sum_auc / query_count,
                "Average R-Precision": sum_rp / query_count,
            }

        else:
            return {"Mean Average Precision": 0.0, "Mean AUC-ROC": 0.0, "Average R-Precision": 0.0}



from logging import getLogger

from .const import DEFAULT_CACHE_FOLDER, MODULE_NAME

_LOGGER = getLogger(MODULE_NAME)


def build_subparser_cache_bed(parser):
    """
    Builds argument parser to support to cache a BED file from local file or BEDbase.
    """
    parser.add_argument("identifier", nargs=1, help="BED file identifier, url, or file path")
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser_cache_bedset(parser):
    """
    Builds argument parser to support to cache a BEDset from local folder or BEDbase.
    """
    parser.add_argument("identifier", nargs=1, help="BED file identifier, url, or file path")
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser_seek(parser):
    """
    Builds argument parser to support to seek the path of BED file or BEDset.
    """
    parser.add_argument("identifier", nargs=1, help="BED file identifier, url, or file path")
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser_inspect(parser):
    """
    Builds argument parser to support to list and count files and subdirectories in the subdirectory bedfiles and bedsets.
    """
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser_cache_tokens(parser):
    """
    Builds argument parser to support to cache tokens from local file or BEDbase.
    """
    parser.add_argument(
        "--bed-id", dest="bed_id", nargs=1, help="Token file identifier, url, or file path"
    )
    parser.add_argument(
        "--universe-id",
        dest="universe_id",
        nargs=1,
        help="Unique identifier for the universe of the token file",
    )
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser_remove(parser):
    """
    Builds argument parser to support to remove bed files or bedsets from the cache folder
    """
    parser.add_argument("identifier", nargs=1, help="BED file identifier, url, or file path")
    parser.add_argument(
        "--cache-folder",
        default=DEFAULT_CACHE_FOLDER,
        help="Cache folder path (default: bed_cache)",
    )

    return parser


def build_subparser(parser):
    """
    Builds argument parser to support the eval command line interface.
    """
    sp = parser.add_subparsers(dest="subcommand")
    msg_by_cmd = {
        "cache-bed": "Cache a BED file from local file or BEDbase",
        "cache-tokens": "Cache tokens from local file or BEDbase",
        "cache-bedset": "Cache a BED set from local folder or BEDbase",
        "seek": "Seek the BED / BEDset path by giving identifier",
        "inspect-bedfiles": "Inspect the contents of bedfile cache folder",
        "inspect-bedsets": "Inspect the contents of bedset cache folder",
        "rm": "Remove the BED/BEDset from cache with given identifier",
    }
    subparsers = {}
    for k, v in msg_by_cmd.items():
        subparsers[k] = sp.add_parser(k, description=v, help=v)
    subparsers["cache-bed"] = build_subparser_cache_bed(subparsers["cache-bed"])
    subparsers["cache-tokens"] = build_subparser_cache_tokens(subparsers["cache-tokens"])
    subparsers["cache-bedset"] = build_subparser_cache_bedset(subparsers["cache-bedset"])
    subparsers["seek"] = build_subparser_seek(subparsers["seek"])
    subparsers["inspect"] = build_subparser_inspect(subparsers["inspect-bedfiles"])
    subparsers["inspect"] = build_subparser_inspect(subparsers["inspect-bedsets"])
    subparsers["rm"] = build_subparser_remove(subparsers["rm"])
    return parser



import argparse
import os
import shlex
import subprocess

from .utils import Timer, time_str


def bedtools_tokenization(
    f: str,
    bedtools_path: str,
    data_folder: str,
    target_folder: str,
    universe: str,
    fraction: float,
) -> None:
    """Uses bedtools to tokenize a raw BED file.

    Args:
        f (str): File name.
        bedtools_path (str): Path to a bedtools binary.
        data_folder (str): The folder where raw BED files reside.
        target_folder (str): The folder that stores tokenized BED files.
        universe (str): Path to a universe file.
        fraction (float): A parameter for bedtools.intersect.
    """
    fname = os.path.join(data_folder, f)
    temp = os.path.join(target_folder, f + "_sorted")
    target = os.path.join(target_folder, f)
    with open(temp, "w") as f_temp:
        subprocess.run(shlex.split(f"sort -k1,1V -k2,2n {fname}"), stdout=f_temp)
    with open(target, "w") as f_target:
        subprocess.run(
            shlex.split(f"{bedtools_path} intersect -a {universe} -b {temp} -u -f {fraction}"),
            stdout=f_target,
        )
    os.remove(temp)


def generate_tokens(
    raw_data_folder: str,
    token_folder: str,
    universe: str,
    file_list: str,
    bedtools: str,
    fraction: float,
) -> None:
    """Tokenizes raw BED files specified by file_list.

    Tokenizes raw BED files specified by file_list. First, checks existing files
    in token_folder. If token_folder has all the tokenized BED files, then does
    nothing. Otherwise, tokenizes raw BED files that are in file_list but not
    in token_folder.

    Args:
        raw_data_folder (str): The folder where raw BED files reside.
        token_folder (str): The folder to store tokenized BED files.
        universe (str): The path to a universe file.
        file_list (str): The path to a file which contains selected BED files per row.
        bedtools (str): The path to a bedtools binary.
        fraction (float): A parameter for bedtools.intersect.
    """
    usize = 0
    with open(universe, "r") as f:
        for _ in f:
            usize += 1
    print(f"\033[93mUniverse size is {usize}\033[00m")

    all_set = []
    with open(file_list, "r") as fin:
        for fname in fin:
            name = fname.strip()
            all_set.append(name)
    all_set = set(all_set)

    if os.path.exists(token_folder):
        files = os.listdir(token_folder)
        existing_set = set([f.strip() for f in files])
        not_covered = all_set - existing_set
        number = len(not_covered)
        if number == 0:
            print(f"Use the existing folder {token_folder}", flush=True)
            return
        else:
            print(
                f"Folder {token_folder} exists with {number} files not processed. Continue...",
                flush=True,
            )
    else:
        os.makedirs(token_folder)
        not_covered = all_set
    for f in not_covered:
        bedtools_tokenization(f, bedtools, raw_data_folder, token_folder, universe, fraction)


def main(args: argparse.Namespace):
    """Generates tokenized BED files.

    Calls generate_tokens using the arguments in args. Prints status
    information.

    Args:
        args (argparse.Namespace): See the definition of the ArgumentParser.
    """
    local_timer = utils.Timer()
    print(f"Entering hard tokenization. Results stored in {args.token_folder}")
    generate_tokens(
        args.data_folder,
        args.token_folder,
        args.universe,
        args.file_list,
        args.bedtools_path,
        args.fraction,
    )
    tokenization_time = local_timer.t()
    print(f"Hard tokenization takes {utils.time_str(tokenization_time)}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-folder",
        type=str,
        default="/scratch/gz5hp/encode3/cell/datasets",
        help="path to the folder that stores BED files",
    )
    parser.add_argument(
        "--file-list",
        type=str,
        default="/home/gz5hp/encode3_proj/all_file_list.txt",
        help="list of BED files that will be tokenized",
    )

    parser.add_argument(
        "--token-folder",
        type=str,
        default="/scratch/gz5hp/encode3/cell/tokens",
        help="folder that stores tokenized files",
    )
    # parameters for hard tokenization
    parser.add_argument(
        "--universe",
        type=str,
        default="/home/gz5hp/encode3_proj/GRCh38-universe.bed",
        help="path to a universe file",
    )
    parser.add_argument(
        "--bedtools-path",
        type=str,
        default="/scratch/gz5hp/genomes/bedtools",
        help="path to the bedtools binary",
    )
    parser.add_argument(
        "--fraction",
        type=float,
        default=1.0e-9,
        help="a parameter for bedtools.intersect",
    )

    args = parser.parse_args()
    if os.path.exists(args.file_list):
        main(args)



from typing import List, Union

import numpy as np


def mean_pooling(region_vectors: List[Union[np.ndarray, None]]) -> np.ndarray:
    """
    Mean pooling of region vectors.

    The function first filters out None values and then computes the mean of the remaining vectors.

    :param region_vectors | None: region vectors
    :return: mean pooled vector
    """
    region_vectors = [rv for rv in region_vectors if rv is not None]
    if len(region_vectors) == 0:
        return None

    region_vectors = np.array(region_vectors)
    return np.mean(region_vectors, axis=0)


def max_pooling(region_vectors: List[Union[np.ndarray, None]]) -> np.ndarray:
    """
    Max pooling of region vectors.

    The function first filters out None values and then computes the max of the remaining vectors.

    :param region_vectors | None: region vectors
    :return: max pooled vector
    """
    region_vectors = [rv for rv in region_vectors if rv is not None]
    if len(region_vectors) == 0:
        return None

    region_vectors = np.array(region_vectors)
    return np.max(region_vectors, axis=0)



import glob
import logging
import os
import random
import select
import shutil
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

import numpy as np

try:
    import torch
except ImportError:
    raise ImportError(
        "Please install Machine Learning dependencies by running 'pip install geniml[ml]'"
    )
from gtars.utils import read_tokens_from_gtok
from yaml import safe_dump, safe_load

if TYPE_CHECKING:
    from gensim.models import Word2Vec as GensimWord2Vec

from ..const import GTOK_EXT
from ..tokenization.main import Tokenizer, TreeTokenizer
from .const import (
    CONFIG_FILE_NAME,
    DEFAULT_EMBEDDING_DIM,
    DEFAULT_EPOCHS,
    DEFAULT_INIT_LR,
    DEFAULT_MIN_COUNT,
    DEFAULT_MIN_LR,
    DEFAULT_WINDOW_SIZE,
    EMBEDDING_DIM_KEY,
    EMBEDDING_DIM_KEY_OLD,
    LR_TYPES,
    MODEL_FILE_NAME,
    MODULE_NAME,
    UNIVERSE_FILE_NAME,
    VOCAB_SIZE_KEY,
)
from .models import Region2Vec

_LOGGER = logging.getLogger(MODULE_NAME)


def prRed(skk: str) -> None:
    """Prints the input string skk in the red font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[91m{skk}\033[00m")


def prGreen(skk: str) -> None:
    """Prints the input string skk in the green font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[92m{skk}\033[00m")


def prYellow(skk: str) -> None:
    """Prints the input string skk in the yellow font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[93m{skk}\033[00m")


def prLightPurple(skk: str) -> None:
    """Prints the input string skk in the light purple font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[94m{skk}\033[00m")


def prPurple(skk: str) -> None:
    """Prints the input string skk in the purple font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[95m{skk}\033[00m")


def prCyan(skk: str) -> None:
    """Prints the input string skk in the cyan font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[96m{skk}\033[00m")


def prLightGray(skk: str) -> None:
    """Prints the input string skk in the gray font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[97m{skk}\033[00m")


def prBlack(skk: str) -> None:
    """Prints the input string skk in the black font.

    Args:
        skk (str): The string to print.
    """
    print(f"\033[98m{skk}\033[00m")


_log_path = None


def set_log_path(path):
    global _log_path
    _log_path = path


class Timer:
    """Records the running time.

    Uses Timer.s() or Timer() to record the start time. Then, calls Timer.t() to get the
    elapsed time in seconds.
    """

    def __init__(self):
        """Initializes a Timer object and starts the timer."""
        self.v = time.time()

    def s(self):
        """Restarts the timer."""
        self.v = time.time()

    def t(self):
        """Gives the elapsed time.

        Returns:
            float: The elapsed time in seconds.
        """
        return time.time() - self.v


def time_str(t: float) -> str:
    """Converts time in float to a readable format.

    Converts time in float to hours, minutes, or seconds based on the value of
    t.

    Args:
        t (float): Time in seconds.

    Returns:
        str: Time in readable time.
    """
    if t >= 3600:
        return f"{t / 3600:.2f}h"
    if t >= 60:
        return f"{t / 60:.2f}m"
    return f"{t:.2f}s"


def timed_response(prompt: str, wait_time: int, default: str):
    """Prints prompt and waits for response.

    Args:
        prompt (str): The question asks for a response.
        wait_time (int): The number of seconds for waiting.
        default (str): If no response received, uses default as the response.

    Returns:
        str: a response given by the user or the default one.
    """
    print(prompt, end="", flush=True)
    i, o, e = select.select([sys.stdin], [], [], wait_time)
    if i:
        ans = sys.stdin.readline().strip()
        if ans not in ["y", "n"]:
            print(f"\033[91m{default}\033[00m")
            return default
        else:
            return ans
    else:
        print(f"\033[91m{default}\033[00m")
        return default


def log(obj: str, filename: str = "log.txt") -> None:
    """Adds information in obj to a file specified by filename.

    Adds information in obj to a file (default: log.txt) and prints obj.

    Args:
        obj (str): A string.
        filename (str, optional): The log file name. Defaults to "log.txt".
    """
    print(obj)
    if _log_path is not None:
        with open(os.path.join(_log_path, filename), "a") as f:
            f.write(obj)
            f.write("\n")


class lr_scheduler:
    """Changes the learning rate.

    Changes the learning rate using the mode of linear or milestones.
    If mode = "linear", then the learning rate linearly decreases after certain
    epochs.
    If mode = "milestones", then the learning rate decreases at specified
    epochs.
    """

    def __init__(
        self,
        init_lr: float,
        end_lr: float,
        epochs: int,
        lr_info: Dict[str, Union[int, float, list]],
        mode: str = "linear",
    ):
        """Initializes the learning rate scheduler.

        Args:
            init_lr (float): The initial learning rate.
            end_lr (float): The last learning rate.
            epochs (int): The number of training epochs.
            lr_info (dict[str,Union[int,list]]): Dictionary storing information
                for learning rate scheduling.
            mode (str, optional): The mode of learning rate scheduling.
                Defaults to "linear".
        """
        self.lr = init_lr
        self.end_lr = end_lr
        self.init_lr = init_lr
        self.mode = mode
        self.epochs = epochs
        self.lr_info = lr_info
        self.count = 0
        if mode == "linear":
            self.freq = lr_info["freq"]

    def step(self):
        """Updates the learning rate.

        Returns:
            float: Current learning rate.
        """
        self.count += 1
        if self.mode == "linear":
            if self.count % self.freq == 0:
                self.lr = self.init_lr - (self.init_lr - self.end_lr) / self.epochs * self.count
        elif self.mode == "milestone":
            milestones = np.array(self.lr_info["milestones"])
            power = (milestones <= self.count).sum()
            self.lr = self.init_lr * np.power(self.lr_info["ratio"], float(power))
            if self.lr < self.end_lr:
                self.lr = self.end_lr
        return self.lr


def ensure_dir(folder: str, default: str = "y") -> None:
    """Makes sure the folder exists.

    Makes sure the folder exists. If the folder exists, then asks the user to
    keep [n] or delete [y] it. If no response received after 5 secs, then
    deletes the folder and create a new one.

    Args:
        folder (str): The folder to be created.
        default (str, optional): Choose whether to delete [y] or keep [n] the
          folder. Defaults to y.
    """
    if os.path.exists(folder):
        if default == "y":
            prompt = f"\033[91m{folder} exists,remove?([y]/n):\033[00m "
        else:
            prompt = f"\033[91m{folder} exists,remove?(y/[n]):\033[00m "
        ans = timed_response(prompt, 5, default)
        if ans != "n":
            shutil.rmtree(folder)
        else:
            return
    os.makedirs(folder, exist_ok=True)


class LearningRateScheduler:
    """
    Simple class to track learning rates of the training procedure

    Based off of: https://machinelearningmastery.com/using-learning-rate-schedules-deep-learning-models-python-keras/
    """

    def __init__(
        self,
        init_lr: float = DEFAULT_INIT_LR,
        min_lr: float = DEFAULT_MIN_LR,
        type: LR_TYPES = "exponential",
        decay: float = None,
        n_epochs: int = None,
    ):
        """
        :param float init_lr: The initial learning rate
        :param float min_lr: The minimum learning rate
        :param str type: The type of learning rate schedule to use. Must be one of ['linear', 'exponential'].
        :param float decay: The decay rate to use. If None, this will be calculated from init_lr and n_epochs.
        :param int n_epochs: The number of epochs to train for. Only used if decay is None.
        """
        self.init_lr = init_lr
        self.min_lr = min_lr
        self.n_epochs = n_epochs

        # convert type to learning rate if necessary
        if type not in ["constant", "linear", "exponential"]:
            raise ValueError(
                f"Unknown schedule type: {type}. Must be one of ['constant', 'linear', 'exponential']."
            )

        self.type = type

        # init the current lr and iteration
        self._current_lr = init_lr
        self._iter = 1

        # init decay rate
        if decay is None:
            _LOGGER.warning(
                "No decay rate provided. Calculating decay rate from init_lr and n_epochs."
            )
            self.decay = init_lr / n_epochs
        else:
            self.decay = decay

    def _update_linear(self, epoch: int):
        """
        Update the learning rate using a linear schedule.

        :param int epoch: The current epoch
        """

        lr = self.init_lr - (self.decay * epoch)
        return max(lr, self.min_lr)

    def _update_exponential(self, epoch: int):
        """
        Update the learning rate using an exponential schedule.

        :param int epoch: The current epoch
        """
        lr = self.get_lr() * (1 / (1 + self.decay * epoch))
        return max(lr, self.min_lr)

    def update(self):
        # update the learning rate according to the type
        if self.type == "linear":
            self._current_lr = self._update_linear(self._iter)
            self._iter += 1
        elif self.type == "exponential":
            self._current_lr = self._update_exponential(self._iter)
            self._iter += 1
        elif self.type == "constant":
            pass  # do nothing
        else:
            raise ValueError(f"Unknown schedule type: {self.type}")

    def get_lr(self):
        return self._current_lr


def shuffle_documents(
    documents: List[List[any]],
    n_shuffles: int = 1,
    threads: int = None,
) -> List[List[any]]:
    """
    Shuffle around the genomic regions for each cell to generate a "context".

    :param List[List[str]] documents: the document list to shuffle.
    :param int n_shuffles: The number of shuffles to conduct.
    """

    def shuffle_list(list: List[any], n: int) -> List[any]:
        for _ in range(n):
            random.shuffle(list)
        return list

    _LOGGER.debug(f"Shuffling documents {n_shuffles} times.")
    shuffled_documents = documents.copy()
    with ThreadPoolExecutor(max_workers=threads) as executor:
        shuffled_documents = list(
            executor.map(
                shuffle_list,
                shuffled_documents,
                [n_shuffles] * len(documents),
            ),
        )
    return shuffled_documents


def export_region2vec_model(
    model: torch.nn.Module,
    tokenizer: Tokenizer,
    path: str,
    checkpoint_file: str = MODEL_FILE_NAME,
    universe_file: str = UNIVERSE_FILE_NAME,
    config_file: str = CONFIG_FILE_NAME,
    **kwargs: Dict[str, any],
):
    """
    Export the region2vec model to a folder

    :param torch.nn.Module model: The model to export
    :param Tokenizer tokenizer: The tokenizer to export
    :param str path: The path to export the model to
    :param str checkpoint_file: The name of the checkpoint file to export
    :param str universe_file: The name of the universe file to export
    :param str config_file: The name of the config file to export
    :param Dict[str, any] kwargs: Any additional arguments to pass to the config file
    """
    # make sure the path exists
    if not os.path.exists(path):
        os.makedirs(path)

    # export the model weights
    torch.save(model.state_dict(), os.path.join(path, checkpoint_file))

    # export the vocabulary
    with open(os.path.join(path, universe_file), "a") as f:
        for region in tokenizer.universe.regions:
            f.write(f"{region.chr}\t{region.start}\t{region.end}\n")

    # export the config (vocab size, embedding size)
    config = {
        VOCAB_SIZE_KEY: len(tokenizer),
        EMBEDDING_DIM_KEY: model.embedding_dim,
    }
    if kwargs:
        config.update(kwargs)

    with open(os.path.join(path, config_file), "w") as f:
        safe_dump(config, f)


def load_local_region2vec_model(
    model_path: str,
    config_path: str,
) -> Tuple[Region2Vec, dict]:
    """
    Load a region2vec model from a local directory

    :param str model_path: The path to the model checkpoint file
    :param str config_path: The path to the model config file
    :param str vocab_path: The path to the model vocabulary file
    """

    # load the model state dict (weights)
    params = torch.load(model_path)

    # get the model config (vocab size, embedding size)
    with open(config_path, "r") as f:
        config = safe_load(f)

    # try with new key first, then old key for backwards compatibility
    embedding_dim = config.get(EMBEDDING_DIM_KEY, config.get(EMBEDDING_DIM_KEY_OLD))
    if embedding_dim is None:
        raise KeyError(
            f"Could not find embedding dimension in config file. Expected key {EMBEDDING_DIM_KEY} or {EMBEDDING_DIM_KEY_OLD}."
        )
    else:
        if EMBEDDING_DIM_KEY_OLD in config:
            _LOGGER.warning(
                f"Found old key {EMBEDDING_DIM_KEY_OLD} in config file. This key will be deprecated in future versions. Please notify this models maintainer."
            )

    model = Region2Vec(
        config[VOCAB_SIZE_KEY],
        embedding_dim=embedding_dim,
    )

    model.load_state_dict(params)

    return model, config


class Region2VecDataset:
    def __init__(
        self,
        data: Union[str, List[str]],
        shuffle: bool = True,
        convert_to_str: bool = False,
    ):
        """
        Initialize a Region2VecDataset.

        The Region2VecDataset is a special dataset that takes advantage of `.gtok` files. These are
        optimized files that contain the tokenized representation of a region set in binary format.
        This allows for much faster loading of the data, and is the recommended way to load data
        for training.

        :param Union[str, List[RegionSet]] data: The data to use for the dataset. This is either a path to a directory container region set files, or a list of region sets.
        :param Tokenizer tokenizer: The tokenizer to use for the dataset.
        :param bool shuffle: Whether or not to shuffle the data before yielding it.
        :param bool convert_to_str: Whether or not to convert the tokens to strings before yielding them.
        """
        self.data = data
        self.shuffle = shuffle
        self.convert_to_str = convert_to_str

        if isinstance(data, str):
            self.data = glob.glob(os.path.join(data, f"*.{GTOK_EXT}"))
        elif isinstance(data, list) and isinstance(data[0], str):
            self.data = data
        else:
            raise ValueError(f"Unknown data type: {type(data)}. Expected str or List[str].")

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        # load the data
        tokens = read_tokens_from_gtok(self.data[idx])

        # shuffle the data if necessary
        if self.shuffle:
            random.shuffle(tokens)

        # return the tokens
        return tokens

    def __iter__(self):
        if len(self) == 0:
            return
        for idx in range(len(self)):
            tokens = self[idx]
            if self.convert_to_str:
                tokens = [str(token) for token in tokens]
            yield tokens

    def __repr__(self):
        return f"Region2VecDataset(data={self.data}, shuffle={self.shuffle})"


def train_region2vec_model(
    dataset: Region2VecDataset,
    embedding_dim: int = DEFAULT_EMBEDDING_DIM,
    window_size: int = DEFAULT_WINDOW_SIZE,
    epochs: int = DEFAULT_EPOCHS,
    min_count: int = DEFAULT_MIN_COUNT,
    num_cpus: int = 1,
    seed: int = 42,
    save_checkpoint_path: str = None,
    gensim_params: dict = {},
    load_from_checkpoint: str = None,
) -> "GensimWord2Vec":
    """
    Train a gensim Word2Vewc model on the given dataset.

    :param Region2VecDataset data: Data to train on. This is a dataset of tokens.
    :param int embedding_dim: Embedding dimension for the model.
    :param int window_size: Window size for the model.
    :param int epochs: Number of epochs to train for.
    :param int min_count: Minimum count for a region to be included in the vocabulary.
    :param int num_cpus: Number of cpus to use for training.
    :param int seed: Seed to use for training.
    :param str save_checkpoint_path: Path to save the model checkpoints to.
    :param dict gensim_params: Additional parameters to pass to the gensim model.
    :param str load_from_checkpoint: Path to a checkpoint to load from.

    :return GensimWord2Vec: The gensim model that was trained.
    """
    # we only need gensim if we are training
    from gensim.models import Word2Vec as GensimWord2Vec
    from gensim.models.callbacks import CallbackAny2Vec

    class TrainingCallback(CallbackAny2Vec):
        """Callback to print loss after each epoch."""

        def __init__(self):
            self.epoch = 0

        def on_epoch_end(self, model: GensimWord2Vec):
            # log the loss
            loss = model.get_latest_training_loss()
            print("Loss after epoch {}: {}".format(self.epoch, loss))
            _LOGGER.info(f"EPOCH {self.epoch} COMPLETE.")
            self.epoch += 1

            # save the model
            if save_checkpoint_path is not None:
                # make sure the path exists
                if not os.path.exists(save_checkpoint_path):
                    os.makedirs(save_checkpoint_path)
                # save the model
                model.save(os.path.join(save_checkpoint_path, f"epoch_{self.epoch}.model"))

    # create gensim model that will be used to train
    if load_from_checkpoint is not None:
        _LOGGER.info(f"Loading model from checkpoint: {load_from_checkpoint}")
        gensim_model = GensimWord2Vec.load(load_from_checkpoint)
    else:
        _LOGGER.info("Creating new gensim model.")
        gensim_model = GensimWord2Vec(
            vector_size=embedding_dim,
            window=window_size,
            min_count=min_count,
            workers=num_cpus,
            seed=seed,
            **gensim_params,
        )
        _LOGGER.info("Building vocabulary.")
        gensim_model.build_vocab(dataset)

    _LOGGER.info("Training model.")
    gensim_model.train(
        dataset,
        epochs=epochs,  # train for 1 epoch at a time, shuffle data each time
        compute_loss=True,
        total_words=gensim_model.corpus_total_words,
        callbacks=[TrainingCallback()],
    )

    _LOGGER.info("Training complete. Moving weights to pytorch model.")
    return gensim_model



import os
import pickle
from typing import Dict, List, Tuple, Union

import numpy as np
from gensim.models import Word2Vec

from ..region2vec import Region2VecExModel


def genome_distance(u: Tuple[int, int], v: Tuple[int, int]) -> float:
    """Computes the genome distance between two regions.

    Assumes that the two regions, u and v, are on the same chromosome.

    Args:
        u (tuple[int, int]): A region denoted by its start and end positions.
        v (tuple[int, int]): A region denoted by its start and end positions.

    Returns:
        float: The genome distance between the two regions.
    """
    return float(u[1] < v[1]) * max(v[0] - u[1] + 1, 0) + float(u[1] >= v[1]) * max(
        u[0] - v[1] + 1, 0
    )


def cosine_distance(x: np.ndarray, y: np.ndarray) -> float:
    """Calculates the cosine distance between two embedding vectors.

    Args:
        x (np.ndarray): An embedding vector.
        y (np.ndarray): An embedding vector.

    Returns:
        float: The cosine distance between two embedding vectors.
    """
    return (1 - ((x / np.linalg.norm(x)) * (y / np.linalg.norm(y))).sum()) / 2


class BaseEmbeddings:
    """Wraps embeddings and the corresponding regions in one object.

    Attributes:
        embeddings (np.ndarray): Region embedding vectors.
        vocab (list[str]): A list of regions in the format of chr:start-end.
    """

    def __init__(self, embeddings, vocab):
        self.embeddings = embeddings
        self.vocab = vocab


def get_bin_embeddings(universe_file: str, tokenized_files: list[str]) -> BaseEmbeddings:
    """Gets a BaseEmbeddings object for binary embeddings.

    Args:
        universe_file (str): The path to a universe file.
        tokenized_files (list[str]): A list of tokoenized BED files (in full
            paths).

    Returns:
        BaseEmbeddings: A BaseEmbeddings object for binary embeddings.
    """
    vocab = []
    with open(universe_file, "r") as f:
        for line in f:
            elements = line.strip().split("\t")
            region = f"{elements[0]}:{elements[1]}-{elements[2]}"
            vocab.append(region)
    vocab_dict = {v: i for i, v in enumerate(vocab)}
    print("vocab size is", len(vocab))
    bin_embeds = np.zeros((len(vocab), len(tokenized_files)))
    for i, token_file in enumerate(tokenized_files):
        with open(token_file, "r") as f:
            for line in f:
                elements = line.strip().split("\t")
                region = f"{elements[0]}:{elements[1]}-{elements[2]}"
                if region in vocab_dict:
                    bin_embeds[vocab_dict[region]][i] = 1
    bin_embed_obj = BaseEmbeddings(bin_embeds, vocab)
    return bin_embed_obj


def get_pca_embeddings(
    bin_embed_obj: BaseEmbeddings, dim: int, kwargs: Dict[str, Union[int, float]] = {}
) -> BaseEmbeddings:
    """Gets PCA embeddings from binary embeddings.

    Args:
        bin_embed_obj (BaseEmbeddings): A BaseEmbeddings object for binary embeddings.
        dim (int): Number of dimensions for PCA embeddings.
        kwargs (dict[str, Union[int, float]], optional): Parameters passed to
            PCA. Defaults to {}.

    Returns:
        BaseEmbeddings: A BaseEmbeddings object for PCA embeddings.
    """
    from sklearn.decomposition import PCA

    embeds = PCA(n_components=dim, **kwargs).fit_transform(bin_embed_obj.embeddings)
    pca_embed_obj = BaseEmbeddings(embeds, bin_embed_obj.vocab)
    return pca_embed_obj


def get_umap_embeddings(
    bin_embed_obj: BaseEmbeddings, dim: int, kwargs: Dict[str, Union[int, float]] = {}
) -> BaseEmbeddings:
    """Gets UMAP embeddings from binary embeddings.

    Args:
        bin_embed_obj (BaseEmbeddings): A BaseEmbeddings object for binary embeddings.
        dim (int): Number of dimensions for UMAP embeddings.
        kwargs (dict[str, Union[int, float]], optional): Parameters passed to
            UMAP. Defaults to {}.

    Returns:
        BaseEmbeddings: A BaseEmbeddings object for UMAP embeddings.
    """
    import umap

    embeds = umap.UMAP(n_components=dim, **kwargs).fit_transform(bin_embed_obj.embeddings)
    umap_embed_obj = BaseEmbeddings(embeds, bin_embed_obj.vocab)
    return umap_embed_obj


def save_base_embeddings(base_embed_obj: BaseEmbeddings, file_name: str) -> None:
    """Saves the BaseEmbeddings object to disk.

    Args:
        base_embed_obj (BaseEmbeddings): A BaseEmbeddings object.
        file_name (str): Save the BaseEmbeddings object to file_name.
    """
    with open(file_name, "wb") as f:
        pickle.dump(base_embed_obj, f)


def load_base_embeddings(path: str) -> Tuple[np.ndarray, List[str]]:
    """Loads a BaseEmbeddings object.

    Args:
        path (str): The path to a BaseEmbeddings object.

    Returns:
        tuple[np.ndarray, list[str]]: Embedding vectors and the corresponding
            region list.
    """
    with open(path, "rb") as f:
        base_embed_obj = pickle.load(f)
    return base_embed_obj.embeddings, base_embed_obj.vocab


def load_genomic_embeddings(
    model_path: str, embed_type: str = "region2vec"
) -> Tuple[np.ndarray, List[str]]:
    """Loads genomic region embeddings based on the type.

    Args:
        model_path (str): The path to a saved model, or a huggingface repo of a model.
        embed_type (str, optional): The model type. Defaults to "region2vec".
            Can be "region2vec", "base", or "huggingface".

    Returns:
        tuple[np.ndarray, list[str]]: Embedding vectors and the corresponding
            region list.
    """
    if os.path.exists(model_path):
        # try to load local
        if embed_type == "region2vec":
            model = Word2Vec.load(model_path)
            regions_r2v = model.wv.index_to_key
            embed_rep = model.wv.vectors
            return embed_rep, regions_r2v
        elif embed_type == "base":
            embed_rep, regions_r2v = load_base_embeddings(model_path)
            return embed_rep, regions_r2v

    else:
        # try to load from huggingface
        exmodel = Region2VecExModel(model_path)
        embed_rep = exmodel.model.projection.weight.data.numpy()
        regions_r2v = [region2vocab_modify(r) for r in exmodel.tokenizer.universe.regions]
        # remove embeddings representing unknown token and padding token
        return embed_rep[:-2], regions_r2v


def region2vocab_modify(region) -> str:
    """Convert a builtins.Region object to a string in the format of chr:start-end.

    Args:
        region (builtins.Region): A region stored in tokenizer

    Returns:
        str: region string in standardized format chr:start-end.
    """
    return f"{region.chr}:{region.start}-{region.end}"


def sort_key(x: str) -> Tuple[int, int]:
    """Extracts chromosome in number and the start position of a region.

    Args:
        x (str): A region in the chr:start-end position.

    Returns:
        tuple[int, int]: Chromosome in number and the start position.
    """
    elements = x.split(":")
    chr_idx = elements[0][3:]
    try:
        idx = int(chr_idx)
    except ValueError:
        idx = 23
        for c in chr_idx:
            idx += ord(c)
    start = int(elements[1].split("-")[0].strip())
    return idx, start


def get_vocab(model_path: str, type: str = "base", ordered: bool = True) -> List[str]:
    """Gets vocab from a model.

    Args:
        model_path (str): The path to a saved model.
        type (str, optional): The embedding type. Defaults to "base".
        ordered (bool, optional): Choose whether to sort the regions. Defaults
            to True.

    Returns:
        list[str]: A list of regions.
    """

    if type == "region2vec":
        model = Word2Vec.load(model_path)
        regions_r2v = model.wv.index_to_key
    elif type == "base":
        _, regions_r2v = load_base_embeddings(model_path)
    if ordered:
        regions_r2v = sorted(regions_r2v, key=sort_key)
    return regions_r2v


def write_vocab(vocab: List[str], file_name: str) -> None:
    """Writes a list of regions to a file.

    Args:
        vocab (list[str]): A list of regions in the format of chr:start-end.
        file_name (str): Saves vocab as file_name.
    """
    with open(file_name, "w") as f:
        for v in vocab:
            elements = v.split(":")
            chr = elements[0].strip()
            s, e = elements[1].split("-")
            s = s.strip()
            e = e.strip()
            f.write(f"{chr}\t{s}\t{e}\n")



from ubiquerg import VersionInHelpParser

from ._version import __version__
from .const import *


def build_argparser(parser: VersionInHelpParser = None) -> VersionInHelpParser:
    """
    Parse command-line arguments passed to the pipeline.

    :param VersionInHelpParser parser: an argument parser object (argparse.ArgumentParser)
    :return: the argument parser object
    """
    # Argument Parsing
    ###########################################################################
    if parser is None:
        parser = VersionInHelpParser(
            prog=PKG_NAME,
            version=__version__,
            description="%(prog)s - embed single-cell data as region vectors",
        )

    # Pipeline-specific arguments
    parser.add_argument(
        "-i",
        "--input",
        default=None,
        type=str,
        required=True,
        help="Path to MarketMatrix format count matrix.",
    )

    parser.add_argument(
        "--nothreads",
        dest="nothreads",
        default=1,
        help="Number of available processors for  " "Word2Vec training.",
    )

    parser.add_argument(
        "--noreads",
        dest="noreads",
        default=2,
        help="Minimum number of reads that overlap a region " "for that region to be included.",
    )

    parser.add_argument(
        "--window-size",
        dest="window_size",
        default=100,
        help="Word2Vec window size.",
    )

    parser.add_argument(
        "--epochs",
        dest="epochs",
        default=30,
        help="Number of epochs for training",
    )

    parser.add_argument(
        "--dimension",
        dest="dimension",
        default=100,
        help="Number of dimensions to train the word2vec " "model.",
    )

    parser.add_argument(
        "--min-count",
        dest="min_count",
        default=10,
        help="Minimum count for Word2Vec model.",
    )

    return parser



import os

import numpy as np
import pytest

from geniml.search.backends import HNSWBackend
from geniml.text2bednn.text2bednn import Vec2VecFNN
from geniml.text2bednn.utils import metadata_dict_from_csv

DATA_FOLDER_PATH = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "tests", "data"
)

np.random.seed(100)


@pytest.fixture
def csv_path():
    """

    Returns: path of the testing metadata csv file

    """
    return os.path.join(DATA_FOLDER_PATH, "ATAC_hg38_sample_for_geniml.csv")


@pytest.fixture
def col_names():
    """
    :return: the columns that are needed metadata csv
    """
    return {
        "tissue",
        "cell_line",
        "tissue_lineage",
        "tissue_description",
        "diagnosis",
        "sample_name",
        "antibody",
    }


@pytest.fixture
def nl_payloads():
    """A list of mock payloads containing natural language metadata and matching BED files"""
    return [
        {"text": "heart muscle", "files": ["b1.bed", "b2.bed", "b3.bed"]},
        {"text": "ipf", "files": ["b1.bed", "b3.bed"]},
        {"text": "healthy control", "files": ["b2.bed"]},
    ]


@pytest.fixture
def bed_payloads():
    """A list of mock payloads containing BED file names"""
    return [
        {"name": "b1.bed"},
        {"name": "b2.bed"},
        {"name": "b3.bed"},
        {"name": "b4.bed"},
        {"name": "b5.bed"},
        {"name": "b6.bed"},
    ]


def test_metadata_dict(csv_path, col_names):
    """
    Test reading metadata from csv file
    """
    metadata_dict1 = metadata_dict_from_csv(csv_path, col_names)
    assert metadata_dict1

    metadata_dict2 = metadata_dict_from_csv(csv_path, col_names, chunk_size=1)
    assert metadata_dict2

    assert metadata_dict1 == metadata_dict2


def test_vec_pair(nl_payloads, bed_payloads, tmp_path_factory):
    """
    Test extracting vector pairs for training/validating from backends
    """
    # mock vectors
    nl_vecs = []
    bed_vecs = []
    for i in range(3):
        nl_vecs.append(np.random.random((3,)))

    for i in range(6):
        bed_vecs.append(np.random.random((2,)))

    temp_data_dir = tmp_path_factory.mktemp("data")
    bed_idx_path = temp_data_dir / "bed_idx.bin"
    nl_idx_path = temp_data_dir / "nl_idx.bin"

    bed_backend = HNSWBackend(local_index_path=str(bed_idx_path), payloads={}, dim=2)
    nl_backend = HNSWBackend(
        local_index_path=str(nl_idx_path),
        payloads={},
        dim=3,
    )

    bed_backend.load(vectors=np.array(bed_vecs), payloads=bed_payloads)
    nl_backend.load(vectors=np.array(nl_vecs), payloads=nl_payloads)


def test_torch_running(tmp_path_factory):
    """
    Test model training
    """
    training_X = np.random.random((90, 1024))
    training_Y = np.random.random((90, 100))
    training_target = np.random.choice([1, -1], size=90)
    validating_X = np.random.random((10, 1024))
    validating_Y = np.random.random((10, 100))
    validating_target = np.random.choice([1, -1], size=10)

    best_embed_folder = tmp_path_factory.mktemp("best_embed")

    v2v_torch1 = Vec2VecFNN()

    v2v_torch1.train(
        training_X,
        training_Y,
        (validating_X, validating_Y),
        save_best=False,
        folder_path=best_embed_folder,
        early_stop=True,
        patience=0.1,
        loss_func="cosine_embedding_loss",
        num_epochs=100,
        batch_size=16,
        num_units=[512, 256],
    )
    v2v_torch1.plot_training_hist(best_embed_folder)
    v2v_torch1.export(best_embed_folder, "v2v_best_epoch.pt")

    v2v_torch2 = Vec2VecFNN()
    v2v_torch2.load_from_disk(
        os.path.join(best_embed_folder, "v2v_best_epoch.pt"),
        os.path.join(best_embed_folder, "config.yaml"),
    )

    input_vecs = np.random.random((5, 1024))

    output1 = v2v_torch1.embedding_to_embedding(input_vecs)
    output2 = v2v_torch2.embedding_to_embedding(input_vecs)

    assert np.array_equal(output1, output2)

    # train the model without validate data
    v2v_torch2.train(training_X, training_Y, num_epochs=10)

    # train the model with contrastiv loss
    v2v_torch_contrast = Vec2VecFNN()

    v2v_torch_contrast.train(
        training_X,
        training_Y,
        (validating_X, validating_Y),
        save_best=False,
        folder_path=best_embed_folder,
        early_stop=True,
        patience=0.1,
        loss_func="cosine_embedding_loss",
        num_epochs=100,
        batch_size=16,
        num_units=[512, 256],
        training_target=training_target,
        validating_target=validating_target,
    )



def build_subparser(parser):
    """Builds an argument parser to support the region2vec command line interface."""
    parser.add_argument("--token-folder", type=str, help="path to tokenized files")
    parser.add_argument("--num-shuffle", type=int, help="number of shufflings/training epochs")
    parser.add_argument("--embed-dim", type=int, help="embedding dimension")
    parser.add_argument("--context-len", type=int, help="Context window size (half)")
    parser.add_argument("--nworkers", type=int, default=10, help="number of workers")
    parser.add_argument(
        "--save-freq",
        type=int,
        default=-1,
        help="Save a model after the given number of training epochs. If -1, then only save the best and latest models",
    )
    parser.add_argument(
        "--save-dir",
        type=str,
        help="path to the folder that saves the training result",
    )
    parser.add_argument(
        "--resume",
        type=str,
        default="",
        help="path to a trained model. If specified, the model will be used to initialize the region2vec embeddings",
    )
    parser.add_argument(
        "--train-alg",
        type=str,
        default="cbow",
        help="training algorithm, select from [cbow, skip-gram]",
    )
    parser.add_argument(
        "--min-count",
        type=int,
        default=5,
        help="threshold for filtering out regions with low frequency in the internal vocabulary",
    )
    parser.add_argument(
        "--neg-samples",
        type=int,
        default=5,
        help="number of noise words in negative sampling, usually between 5-20",
    )
    parser.add_argument("--init-lr", type=float, default=0.1, help="initial learning rate")
    parser.add_argument("--milestones", nargs="+", type=int, default=[100, 200])
    parser.add_argument(
        "--lr-mode",
        type=str,
        default="linear",
        choices=["milestone", "linear"],
        help="type of learning rate scheduler, milestone or linear",
    )
    parser.add_argument(
        "--update-vocab",
        type=str,
        default="once",
        help="[every] update at every epoch; [once] Update once since the vocabulary does not change",
    )
    parser.add_argument("--min-lr", type=float, default=1.0e-6, help="minimum learning rate")
    parser.add_argument("--seed", type=int, default=42, help="random seed")

    return parser



import os
import pickle

os.environ["OPENBLAS_NUM_THREADS"] = "1"

from typing import List, Tuple

import numpy as np
from sklearn.neighbors import NearestNeighbors

from .const import CTT_QUANTILE_MAX, CTT_QUANTILE_MIN, CTT_TEST_RATIO
from .utils import load_genomic_embeddings


def get_ctt_score(
    path: str,
    embed_type: str,
    seed: int = 42,
    num_data: int = 10000,
    num_workers: int = 10,
) -> float:
    """Runs the cluster tendency test (CTT) on a model.

    Args:
        path (str): The path to a model.
        embed_type (str): The type of the model: "region2vec" or "base".
        seed (int, optional): Random seed. Defaults to 42.
        num_data (int, optional): Number of embeddings used for evaluation.
            Defaults to 10000.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Raises:
        ValueError: The number of samples is too small.
        ZeroDivisionError: The denominator of the CTT score is zero.

    Returns:
        float: The CTT score for the model.
    """
    np.random.seed(seed)
    data, vocab = load_genomic_embeddings(path, embed_type)
    num_ori, dimension = data.shape
    if num_data < num_ori:
        num = num_data
    else:
        num = num_ori
    data = data[np.random.choice(num_ori, num)]
    if num < 100:
        raise ValueError(f"Number of samples ({num}) is too small")
    num_samples = int(num * CTT_TEST_RATIO)

    sel_indexes = np.random.choice(num, num_samples)
    data_sample = data[sel_indexes]
    neigh = NearestNeighbors(n_neighbors=2, n_jobs=num_workers).fit(data)
    sample_dist, _ = neigh.kneighbors(data_sample)
    sample_dist_to_nn = sample_dist[:, 1]

    max_vals = np.quantile(data, CTT_QUANTILE_MAX, axis=0)
    min_vals = np.quantile(data, CTT_QUANTILE_MIN, axis=0)
    random_points = np.random.uniform(min_vals, max_vals, (num_samples, dimension))

    random_dist, _ = neigh.kneighbors(random_points, n_neighbors=1)

    random_dist_to_nn = random_dist[:, 0]

    x = sum(sample_dist_to_nn**2)
    y = sum(random_dist_to_nn**2)

    if x + y == 0:
        raise ZeroDivisionError("The denominator is zero")

    return y / (x + y)


def get_ctt_batch(
    batch: List[Tuple[str, str]],
    seed: int = 42,
    num_data: int = 10000,
    save_path: str = None,
    num_workers: int = 10,
) -> List[Tuple[str, float]]:
    """Runs the cluster tendency test (CTT) on a batch of models.

    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type)
            tuples. Model type could be "region2vec" or "base".
        seed (int, optional): Random seed. Defaults to 42.
        num_data (int, optional): Number of embeddings used for evaluation.
            Defaults to 10000.
        save_path (str, optional): Save the results to save_path. Defaults to
            None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        List[Tuple[str, float]]: A list of (model path, CTT score) tuples.
    """
    ctt_arr = []
    for path, embed_type in batch:
        ctt = get_ctt_score(path, embed_type, seed, num_data, num_workers)
        # print(f"{'/'.join(path.split('/')[-3:])}: {ctt:.4f}")
        ctt_arr.append((path, ctt))
    if save_path:
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        with open(save_path, "wb") as f:
            pickle.dump(ctt_arr, f)
    return ctt_arr


def ctt_eval(
    batch: List[Tuple[str, str]],
    num_runs: int = 20,
    num_data: int = 10000,
    save_folder: str = None,
    num_workers: int = 10,
) -> List[Tuple[str, list[float]]]:
    """Runs the CTT on a batch of models for multiple times.

    Runs the cluster tendency test (CTT) for a batch of models for num_runs
    times with different random seeds.


    Args:
        batch (list[tuple[str, str]]): A list of (model path, model type)
            tuples. Model type could be "region2vec" or "base".
        num_runs (int, optional): Number of runs. Defaults to 20.
        num_data (int, optional): Number of embeddings used for evaluation.
            Defaults to 10000.
        save_folder (str, optional): Folder to save the results from each run.
            Defaults to None.
        num_workers (int, optional): Number of parallel processes used.
            Defaults to 10.

    Returns:
        list[tuple[str, list[float]]]: A list of (model path, CTT scores from
            num_runs) tuples.
    """
    results_seeds = []
    for seed in range(num_runs):
        print(f"----------------Run {seed}----------------")
        save_path = os.path.join(save_folder, f"ctt_eval_seed{seed}") if save_folder else None
        result_list = get_ctt_batch(batch, seed, num_data, save_path, num_workers)
        results_seeds.append(result_list)

    ctt_res = [[] for i in range(len(batch))]
    for results in results_seeds:
        for i, res in enumerate(results):
            ctt_res[i].append(res[1])
            assert res[0] == batch[i][0], "key == batch[i][0]"

    mean_ctt = [np.array(r).mean() for r in ctt_res]
    std_ctt = [np.array(r).std() for r in ctt_res]
    # models = [t[0] for t in batch]
    for i in range(len(mean_ctt)):
        print(f"{batch[i][0]}\n CTT (std): {mean_ctt[i]:.4f} ({std_ctt[i]:.4f}) \n")
    ctt_arr = [(batch[i][0], ctt_res[i]) for i in range(len(batch))]
    return ctt_arr
